#ifndef __DATAHANDLE_H__
#define __DATAHANDLE_H__

#include "../std.h"

#include <globus_common.h>
#include <globus_ftp_client.h>
#include <globus_ftp_client_plugin.h>
#include <globus_gass_copy.h>

#include "../misc/condition.h"
#include "../misc/globus_modules.h"

#define MAX_PARALLEL_STREAMS 20
#define MAX_BLOCK_SIZE (1024*1024)

class DataPoint;
class DataBufferPar;
class DataCallback;
class DataPoint::FileInfo;

#ifndef SKIP_HTTPG
class httpg_info_t;
#endif


/// DataHandle is kind of generalized file handle. Differently 
/// from file handle it does not support operations read() and 
/// write(). Instead it initiates operation and uses object of 
/// class DataBufferPar to pass actual data. It also provides 
/// other operations like querying parameters of remote object. 
/// It is used by higher-level classes DataMove and DataMovePar 
/// to provide data transfer service for application.
class DataHandle {
 public:
  /// Reason of transfer failure
  typedef enum {
    common_failure = 0,
    credentials_expired_failure = 1
  } failure_reason_t;
 private:
  typedef enum {
    url_is_unknown,
    url_is_channel,
    url_is_file,
    url_is_ftp,
#ifdef SKIP_HTTPG
    url_is_http,
#else
    url_is_httpg,
#endif
  } url_proto_t;
  /* set by init_handle */
  url_proto_t url_proto;
  DataPoint* url; /* full url with locations, etc. */
  DataBufferPar* buffer; /* pointer to buffer for data */
  string c_url;   /* current url with options stripped */ 
  bool cacheable;
  bool linkable;
  bool is_secure;
  bool force_secure;
  bool force_passive;
  bool reading;
  bool writing;
  bool no_checks;
  /* true if allowed to produce scattered data (non-sequetial offsets) */
  bool allow_out_of_order;
  unsigned long long int range_start;
  unsigned long long int range_end;

  /* ftp */
  GlobusModuleFTPClient ftp_mod;
  bool ftp_active;
  globus_ftp_client_handle_t ftp_handle;
  globus_ftp_client_operationattr_t ftp_opattr;
  globus_ftp_client_plugin_t ftp_plugin;
  globus_thread_t ftp_control_thread;
  int ftp_threads;
  Condition<int> ftp_completed;
  LockSimple ftp_order;
  bool ftp_eof_flag;
  CounterSimple ftp_counter;
  static void ftp_complete_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error);
  static void ftp_get_complete_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error);
  static void ftp_put_complete_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error);
  static void ftp_read_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof);
  static void ftp_check_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof);
  static void ftp_write_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof);
  static void* DataHandle::ftp_read_thread(void *arg);
  static void* DataHandle::ftp_write_thread(void *arg);
  bool start_reading_ftp(DataBufferPar &buf);
  bool start_writing_ftp(DataBufferPar &buf);
  bool stop_reading_ftp(void);
  bool stop_writing_ftp(void);
  bool check_ftp(void);
  bool remove_ftp(void);
  string ftp_dir_path;
  bool mkdir_ftp();
  char ftp_buf[16];
  bool list_files_ftp(list<DataPoint::FileInfo> &files,bool resolve = true);

  /* http */
#ifdef SKIP_HTTPG
  GlobusModuleGASSTransfer http_mod;
#endif
  bool http_active;
  globus_gass_transfer_request_t http_handle;
  bool http_handle_was_initialized;
  unsigned long long int http_offset;
  bool start_reading_http(DataBufferPar &buffer);
  bool start_writing_http(DataBufferPar &buffer);
  bool stop_reading_http(void);
  bool stop_writing_http(void);
  bool check_http(void);
  static void read_http(void* arg,globus_gass_transfer_request_t handle,globus_byte_t* buffer, globus_size_t length, globus_bool_t last_data);
  static void write_http(void* arg,globus_gass_transfer_request_t handle,globus_byte_t* buffer, globus_size_t length, globus_bool_t last_data);
  static void http_callback(void* arg,globus_gass_transfer_request_t handle);
  static void http_check_callback(void* arg,globus_gass_transfer_request_t handle);

#ifndef SKIP_HTTPG
  /* httpg */
  httpg_info_t* httpg_stat;
  bool start_reading_httpg(DataBufferPar &buffer);
  bool stop_reading_httpg(void);
  bool start_writing_httpg(DataBufferPar &buffer);
  bool stop_writing_httpg(void);
  bool remove_httpg(void);
  bool check_httpg(void);
  bool list_files_httpg(list<DataPoint::FileInfo> &files,bool resolve = true);
#endif

  /* bbftp - no bbftp is not supported */
  bool bbftp_active;

  /* file */
  int fd;
  pthread_t file_thread;
  pthread_attr_t file_thread_attr;
  bool start_reading_file(DataBufferPar &buffer);
  bool start_writing_file(DataBufferPar &buffer,DataCallback *space_cb);
  bool stop_reading_file(void);
  bool stop_writing_file(void);
  bool check_file(void);
  bool remove_file(void);
  static void* read_file(void* arg);
  static void* write_file(void* arg);
  CondSimple file_thread_exited;
  bool list_files_file(list<DataPoint::FileInfo> &files,bool resolve = true);

  failure_reason_t failure_code;

  bool init_handle(void); 
  bool deinit_handle(void); 
 public:
  /// Constructor
  /// \param url_ URL. Should not be destroyed before DataHandle itself. 
  DataHandle(DataPoint* url_);
  /// Destructor. No comments.
  ~DataHandle(void);
  /// Start reading data from URL.
  /// Separate thread to transfer data will be created. No other operation
  /// can be performed while 'reading' is in progress.
  /// \param buffer operation will use this buffer to put information into. Should not be destroyed before stop_reading was called and returned.
  /// Returns true on success. 
  bool start_reading(DataBufferPar &buffer);
  /// Start writing data to URL.
  /// Separate thread to transfer data will be created. No other operation 
  /// can be performed while 'writing' is in progress.
  /// \param buffer operation will use this buffer to get information from. Should not be destroyed before stop_writing was called and returned.
  /// space_cb callback which is called if there is not enough to space storing data. Currently implemented only for file:/// URL.
  /// Returns true on success.
  bool start_writing(DataBufferPar &buffer,DataCallback *space_cb = NULL);
  /// Stop reading. It MUST be called after corressponding start_reading 
  /// method. Either after whole data is transfered or to cancel transfer.
  /// Use 'buffer' object to find out when data is transfered.
  bool stop_reading(void);
  /// Same as stop_reading but for corresponding start_writing.
  bool stop_writing(void);
  /// Structure used in analyze() call. 
  /// \param bufsize returns suggested size of buffers to store data.
  /// \param bufnum returns suggested number of buffers.
  /// \param cache returns true if url is allowed to be cached.
  /// \param local return true if URL is accessed locally (file://)
  class analyze_t {
   public:
    long int bufsize;
    int bufnum;
    bool cache;
    bool local;
    bool readonly;
    analyze_t(void):bufsize(-1),bufnum(1),cache(true),local(false),readonly(true) { };
  };
  ///  Analyze url and provide hints.
  /// \param arg returns suggested values.
  bool analyze(analyze_t &arg);
  /// Query remote server or local file system to check if object is
  /// accessible.
  bool check(void);
  /// Remove/delete object at URL.
  bool remove(void);
  /// List files in directory (URL must point to directory/group).
  /// \param files will contain list of file names and optionally their attributes.
  /// \param resolve if false no information about attributes will be retrieved.
  bool list_files(list<DataPoint::FileInfo> &files,bool resolve = true);
  /// Returns true if URL can accept scatterd data (like arbitrary access to local file) for 'writing' operation.
  bool out_of_order(void);
  /// Allow/disallow DataHandle to produce scattered data during 'reading' operation.
  /// \param v true if allowed.
  void out_of_order(bool v);
  /// Allow/disallow to make check for existance of remote file (and probably other checks too) before initiating 'reading' and 'writing' operations.
  /// \param v true if allowed (default is true).
  void additional_checks(bool v) { no_checks=!v; };
  /// Check if additional checks before 'reading' and 'writing' will be performed.
  bool additional_checks(void) { return !no_checks; };
  /// Allow/disallow heavy security during data transfer.
  /// \param v true if allowed (default is true only for gsiftp://).
  void secure(bool v);
  /// Check if heavy security during data transfer is allowed.
  bool secure(void);
  /// Request passive transfers for FTP-like protocols.
  /// \param true to request.
  void passive(bool v);
  /// Returns reason of transfer failure.
  failure_reason_t failure_reason(void) { return failure_code; };
  /// Set range of bytes to retrieve. Default values correspond to whole file. 
  void range(unsigned long long int start = 0,unsigned long long end = 0) { range_start=start; range_end=end; };
};

#endif
