/*
 routines to work with informational files
*/
#include "../std.h"
#include <fstream>
#include "../misc/delete.h"
#include "../misc/stringtoint.h"
#include "../misc/inttostring.h"
#include "../misc/escaped.h"
#include "../misc/log_time.h"
#include "info_types.h"
#include "info_files.h"

const char * const sfx_failed      = ".failed";
const char * const sfx_cancel      = ".cancel";
const char * const sfx_restart     = ".restart";
const char * const sfx_clean       = ".clean";
const char * const sfx_status      = ".status";
const char * const sfx_local       = ".local";
const char * const sfx_errors      = ".errors";
const char * const sfx_rsl         = ".description";
const char * const sfx_diag        = ".diag";
const char * const sfx_lrmsoutput  = ".comment";
const char * const sfx_diskusage   = ".disk";
const char * const sfx_acl         = ".acl";

bool fix_file_permissions(const string &fname,bool executable) {
  mode_t mode = S_IRUSR | S_IWUSR;
  if(executable) { mode |= S_IXUSR; };
  return (chmod(fname.c_str(),mode) == 0);
}

bool fix_file_owner(const string &fname,const JobUser &user) {
  if(getuid() == 0) {
    if(lchown(fname.c_str(),user.get_uid(),user.get_gid()) == -1) {
      olog << "Failed setting file owner: " << fname << endl;
      return false;
    };
  };
  return true;
}

bool fix_file_owner(const string &fname,const JobDescription &desc,const JobUser &user) {
  if(getuid() == 0) {
    uid_t uid = desc.get_uid();
    gid_t gid = desc.get_gid();
    if( uid == 0 ) {
      uid=user.get_uid(); gid=user.get_gid();
    };
    if(lchown(fname.c_str(),uid,gid) == -1) {
      olog << "Failed setting file owner: " << fname << endl;
      return false;
    };
  };
  return true;
}

bool check_file_owner(const string &fname,const JobUser &user) {
  uid_t uid;
  gid_t gid;
  time_t t;
  return check_file_owner(fname,user,uid,gid,t);
}

bool check_file_owner(const string &fname,const JobUser &user,uid_t &uid,gid_t &gid) {
  time_t t;
  return check_file_owner(fname,user,uid,gid,t);
}

bool check_file_owner(const string &fname,const JobUser &user,uid_t &uid,gid_t &gid,time_t &t) {
  struct stat st;
  if(lstat(fname.c_str(),&st) != 0) return false;
  if(!S_ISREG(st.st_mode)) return false;
  uid=st.st_uid; gid=st.st_gid; t=st.st_ctime;
  /* superuser can't run jobs */
  if(uid == 0) return false;
  /* accept any file if superuser */
  if(user.get_uid() != 0) {
    if(uid != user.get_uid()) return false;
  };
  return true;
}

bool job_lrms_mark_put(const JobDescription &desc,JobUser &user,int code) {
  LRMSResult r(code);
  return job_lrms_mark_put(desc,user,r);
}

bool job_lrms_mark_put(const JobDescription &desc,JobUser &user,LRMSResult r) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + ".lrms_done";
  string content = inttostring(r.code());
  content+=" ";
  content+=r.description();
  return job_mark_write_s(fname,content) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_lrms_mark_check(JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + ".lrms_done";
  return job_mark_check(fname);
}

bool job_lrms_mark_remove(JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + ".lrms_done";
  return job_mark_remove(fname);
}

LRMSResult job_lrms_mark_read(JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + ".lrms_done";
  LRMSResult r("-1 Internal error");
  ifstream f(fname.c_str()); if(! f.is_open() ) return r;
  f>>r;
  return r;
}

bool job_cancel_mark_put(const JobDescription &desc,JobUser &user) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + sfx_cancel;
  return job_mark_put(fname) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_cancel_mark_check(JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_cancel;
  return job_mark_check(fname);
}

bool job_cancel_mark_remove(JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_cancel;
  return job_mark_remove(fname);
}

bool job_restart_mark_put(const JobDescription &desc,JobUser &user) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + sfx_restart;
  return job_mark_put(fname) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_restart_mark_check(JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_restart;
  return job_mark_check(fname);
}

bool job_restart_mark_remove(JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_restart;
  return job_mark_remove(fname);
}

bool job_clean_mark_put(const JobDescription &desc,JobUser &user) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + sfx_clean;
  return job_mark_put(fname) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_clean_mark_check(JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_clean;
  return job_mark_check(fname);
}

bool job_clean_mark_remove(JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_clean;
  return job_mark_remove(fname);
}

bool job_errors_mark_put(const JobDescription &desc,JobUser &user) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + sfx_errors;
  return job_mark_put(fname) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_failed_mark_put(const JobDescription &desc,JobUser &user,const string &content) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + sfx_failed;
  if(job_mark_size(fname) > 0) return true;
  return job_mark_write_s(fname,content) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_failed_mark_add(const JobDescription &desc,JobUser &user,const string &content) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + sfx_failed;
  return job_mark_add_s(fname,content) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_failed_mark_check(const JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_failed;
  return job_mark_check(fname);
}

bool job_failed_mark_remove(const JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_failed;
  return job_mark_remove(fname);
}

string job_failed_mark_read(const JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_failed;
  return job_mark_read_s(fname);
}

bool job_diagnostics_mark_put(const JobDescription &desc,JobUser &user) {
  string fname = desc.SessionDir() + sfx_diag;
  return job_mark_put(fname) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_lrmsoutput_mark_put(const JobDescription &desc,JobUser &user) {
  string fname = desc.SessionDir() + sfx_lrmsoutput;
  return job_mark_put(fname) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_diagnostics_mark_add(const JobDescription &desc,JobUser &user,const string &content) {
  string fname = desc.SessionDir() + sfx_diag;
  return job_mark_add_s(fname,content) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_diagnostics_mark_remove(const JobId &id,JobUser &user) {
  string fname1 = user.SessionRoot() + "/" + id + sfx_diag;
  string fname2 = user.ControlDir() + "/job." + id + sfx_diag;
  bool res1 = job_mark_remove(fname1);
  bool res2 = job_mark_remove(fname2);
  return res1 | res2;
}

bool job_lrmsoutput_mark_remove(const JobId &id,JobUser &user) {
  string fname = user.SessionRoot() + "/" + id + sfx_lrmsoutput;
  return job_mark_remove(fname);
}

bool job_diagnostics_mark_move(const JobDescription &desc,JobUser &user) {
  string fname1 = user.SessionRoot() + "/" + desc.get_id() + sfx_diag;
  string fname2 = user.ControlDir() + "/job." + desc.get_id() + sfx_diag;
  int h1=open(fname1.c_str(),O_RDONLY);
  if(h1==-1) return false;
  int h2=open(fname2.c_str(),O_WRONLY | O_CREAT | O_TRUNC,S_IRUSR | S_IWUSR);
  if(h2==-1) { close(h1); return false; };
  fix_file_owner(fname2,desc,user);
  fix_file_permissions(fname2);
  char buf[256];
  int l;
  for(;;) {
    l=read(h1,buf,256);
    if((l==0) || (l==-1)) break;
    write(h2,buf,l);
  };
  close(h1); close(h2);
  unlink(fname1.c_str());
  return true;
}

bool job_lrmsoutput_mark_move(const JobDescription &desc,JobUser &user) {
  /* take 1 last non-empty line from lrms output file and put it to failed */
  string fname = desc.SessionDir() + sfx_lrmsoutput;
  char buf[256];
  bool res = true;
  string s;
  s="";
  ifstream f(fname.c_str());
  if(f.is_open()) {
    for(;!f.eof() && f;) {
      istream_readline(f,buf,sizeof(buf));
      if(buf[0]) s=buf;
    };
    f.close();
    if(s.length()) {
      s="\n"+s;
      res = job_failed_mark_add(desc,user,s);
    };
  };
  unlink(fname.c_str());
  return res;
}

bool job_lrmsoutput_mark_get(JobDescription &desc,JobUser &user) {
  /* take 1 last non-empty line from lrms output file and put it to failed */
  string fname = desc.SessionDir() + sfx_lrmsoutput;
  char buf[256];
  bool res = true;
  string s("");
  ifstream f(fname.c_str());
  if(f.is_open()) {
    for(;!f.eof() && f;) {
      istream_readline(f,buf,sizeof(buf));
      if(buf[0]) s=buf;
    };
    f.close();
    if(s.length()) desc.AddFailure(s);
  };
  unlink(fname.c_str());
  return res;
}

bool job_stdlog_move(const JobDescription &desc,JobUser &user,const string &logname) {
/*

 Status data is now available during runtime.

  string fname = desc.SessionDir() + "/" + logname;
  string fname_err = user.ControlDir() + "/job." + desc.get_id() + sfx_errors;
  string fname_diag = user.ControlDir() + "/job." + desc.get_id() + sfx_diag;
  int h2=open(fname.c_str(),O_WRONLY | O_CREAT | O_APPEND,S_IRUSR | S_IWUSR);
  if(h2==-1) { return false; };
  char buf[256];
  int l;
  int h1=open(fname_err.c_str(),O_RDONLY);
  if(h1==-1) { close(h2); return false; };
  fix_file_owner(fname,desc,user);
  for(;;) {
    l=read(h1,buf,256);
    if((l==0) || (l==-1)) break;
    write(h2,buf,l);
  };
  close(h1);
  write(h2,"\n----------------------------------------\n",42);
  h1=open(fname_diag.c_str(),O_RDONLY);
  if(h1==-1) { close(h2); return false; };
  for(;;) {
    l=read(h1,buf,256);
    if((l==0) || (l==-1)) break;
    write(h2,buf,l);
  };
  close(h1); 
  close(h2);
*/
  return true;
}

bool job_mark_put(const string &fname) {
  int h=open(fname.c_str(),O_WRONLY | O_CREAT,S_IRUSR | S_IWUSR);
  if(h==-1) return false; close(h); return true;
}

bool job_mark_check(const string &fname) {
  struct stat st;
  if(lstat(fname.c_str(),&st) != 0) return false;
  if(!S_ISREG(st.st_mode)) return false;
  return true;
}

bool job_mark_remove(const string &fname) {
  if(unlink(fname.c_str()) != 0) { if(errno != ENOENT) return false; };
  return true;
}

string job_mark_read_s(const string &fname) {
  string s("");
  ifstream f(fname.c_str()); if(! f.is_open() ) return s;
  char buf[256]; f.getline(buf,254); s=buf;
  return s;
}

long int job_mark_read_i(const string &fname) {
  ifstream f(fname.c_str()); if(! f.is_open() ) return -1;
  char buf[32]; f.getline(buf,30); f.close();
  char* e; long int i;
  i=strtol(buf,&e,10); if((*e) == 0) return i;
  return -1;
}

bool job_mark_write_s(const string &fname,const string &content) {
  int h=open(fname.c_str(),O_WRONLY | O_CREAT | O_TRUNC,S_IRUSR | S_IWUSR);
  if(h==-1) return false;
  write(h,(const void *)content.c_str(),content.length());
  close(h); return true;
}

bool job_mark_add_s(const string &fname,const string &content) {
  int h=open(fname.c_str(),O_WRONLY | O_CREAT | O_APPEND,S_IRUSR | S_IWUSR);
  if(h==-1) return false;
  write(h,(const void *)content.c_str(),content.length());
  close(h); return true;
}

time_t job_mark_time(const string &fname) {
  struct stat st;
  if(lstat(fname.c_str(),&st) != 0) return 0;
  return st.st_mtime;
}

long int job_mark_size(const string &fname) {
  struct stat64 st;
  if(lstat64(fname.c_str(),&st) != 0) return 0;
  if(!S_ISREG(st.st_mode)) return 0;
  return st.st_size;
}

bool job_diskusage_create_file(const JobDescription &desc,JobUser &user,unsigned long long int &requested) {
  string fname = desc.SessionDir() + sfx_diskusage;
  int h=open(fname.c_str(),O_WRONLY | O_CREAT,S_IRUSR | S_IWUSR);
  if(h==-1) return false;
  char content[200];
  sprintf(content,"%llu 0\n",requested);
  write(h,content,strlen(content));
  close(h); return true;
}

bool job_diskusage_read_file(const JobDescription &desc,JobUser &user,unsigned long long int &requested,unsigned long long int &used) {
  string fname = desc.SessionDir() + sfx_diskusage;
  int h=open(fname.c_str(),O_RDONLY);
  if(h==-1) return false;
  char content[200];
  ssize_t l=read(h,content,199);
  if(l==-1) { close(h); return false; };
  content[l]=0;
  unsigned long long int req_,use_;
  if(sscanf(content,"%llu %llu",&req_,&use_) != 2) { close(h); return false; };
  requested=req_; used=use_;  
  close(h); return true;
}

bool job_diskusage_change_file(const JobDescription &desc,JobUser &user,signed long long int used,bool &result) {
  // lock file, read, write, unlock
  string fname = desc.SessionDir() + sfx_diskusage;
  int h=open(fname.c_str(),O_RDWR);
  if(h==-1) return false;
  struct flock lock;
  lock.l_type=F_WRLCK;
  lock.l_whence=SEEK_SET;
  lock.l_start=0;
  lock.l_len=0;
  for(;;) {
    if(fcntl(h,F_SETLKW,&lock) != -1) break;
    if(errno == EINTR) continue;
    close(h); return false;
  };
  char content[200];
  ssize_t l=read(h,content,199);
  if(l==-1) {
    lock.l_whence=SEEK_SET; lock.l_start=0; lock.l_len=0;
    lock.l_type=F_UNLCK; fcntl(h,F_SETLK,&lock);
    close(h); return false;
  };
  content[l]=0;
  unsigned long long int req_,use_;
  if(sscanf(content,"%llu %llu",&req_,&use_) != 2) {
    lock.l_whence=SEEK_SET; lock.l_start=0; lock.l_len=0;
    lock.l_type=F_UNLCK; fcntl(h,F_SETLK,&lock);
    close(h); return false;
  };
  if((-used) > use_) {
    result=true; use_=0; // ??????
  }
  else {
    use_+=used; result=true;
    if(use_>req_) result=false;
  };
  lseek(h,0,SEEK_SET);  
  sprintf(content,"%llu %llu\n",req_,use_);
  write(h,content,strlen(content));
  lock.l_whence=SEEK_SET; lock.l_start=0; lock.l_len=0;
  lock.l_type=F_UNLCK; fcntl(h,F_SETLK,&lock);
  close(h); return true;
}

bool job_diskusage_remove_file(const JobDescription &desc,JobUser &user) {
  string fname = desc.SessionDir() + sfx_diskusage;
  return job_mark_remove(fname);
}

time_t job_state_time(const JobId &id,JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_status;
  return job_mark_time(fname);
}

job_state_t job_state_read_file(const JobId &id,const JobUser &user) {
  string fname = user.ControlDir() + "/job." + id + sfx_status;
  bool pending;
  return job_state_read_file(fname,pending);
}

job_state_t job_state_read_file(const JobId &id,const JobUser &user,bool& pending) {
  string fname = user.ControlDir() + "/job." + id + sfx_status;
  return job_state_read_file(fname,pending);
}

bool job_state_write_file(const JobDescription &desc,JobUser &user,job_state_t state,bool pending) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + sfx_status;
  return job_state_write_file(fname,state,pending) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

job_state_t job_state_read_file(const string &fname,bool &pending) {
  ifstream f(fname.c_str());
  if(!f.is_open() ) return JOB_STATE_UNDEFINED; /* can't open file */
  char buf[32];
  f.getline(buf,30);
  /* interpret information */
  char* p = buf;
  if(!strncmp("PENDING:",p,8)) { p+=8; pending=true; } else { pending=false; };
  for(int i = 0;states_all[i].name != NULL;i++) {
    if(!strcmp(states_all[i].name,p)) {
      f.close();
      return states_all[i].id;
    };
  };
  f.close();
  return JOB_STATE_UNDEFINED; /* broken file */
}

bool job_state_write_file(const string &fname,job_state_t state,bool pending) {
  ofstream f(fname.c_str(),ios::out | ios::trunc);
  if(! f.is_open() ) return false; /* can't open file */
  if(pending) f<<"PENDING:";
  f<<states_all[state].name<<endl;
  f.close();
  return true;
}

bool job_description_read_file(JobId &id,JobUser &user,string &rsl) {
  string fname = user.ControlDir() + "/job." + id + sfx_rsl;
  return job_description_read_file(fname,rsl);
}

bool job_description_read_file(const string &fname,string &rsl) {
  char buf[256];
  int n=0;
  ifstream f(fname.c_str());
  if(! f.is_open() ) return false; /* can't open file */
  rsl.erase();
  while(!f.eof()) {
    memset(buf,0,256);
    f.read(buf,255); rsl+=buf;
    for(;(n=rsl.find('\n',n)) != string::npos;) rsl.erase(n,1);
    n=rsl.length();
  };
  f.close();
  return true;
}

bool job_description_write_file(const string &fname,string &rsl) {
  return job_description_write_file(fname,rsl.c_str());
}

bool job_description_write_file(const string &fname,const char* rsl) {
  ofstream f(fname.c_str(),ios::out | ios::trunc);
  if(! f.is_open() ) return false; /* can't open file */
  f.write(rsl,strlen(rsl));
  f.close();
  return true;
}

bool job_acl_read_file(JobId &id,JobUser &user,string &acl) {
  string fname = user.ControlDir() + "/job." + id + sfx_acl;
  return job_description_read_file(fname,acl);
}

bool job_acl_write_file(JobId &id,JobUser &user,string &acl) {
  string fname = user.ControlDir() + "/job." + id + sfx_acl;
  return job_description_write_file(fname,acl.c_str());
}

bool job_local_write_file(const JobDescription &desc,const JobUser &user,JobLocalDescription &job_desc) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + sfx_local;
  return job_local_write_file(fname,job_desc) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

inline void write_pair(ofstream &f,const string &name,const string &value) {
  if(value.length()) f << name << '=' << value << endl;
}

inline void write_pair(ofstream &f,const string &name,const mds_time &value) {
  if(value.defined()) f << name << '=' << value << endl;
}

bool job_local_write_file(const string &fname,JobLocalDescription &job_desc) {
  ofstream f(fname.c_str(),ios::out | ios::trunc);
  if(! f.is_open() ) return false; /* can't open file */
  write_pair(f,"jobreport",job_desc.jobreport);
  write_pair(f,"lrms",job_desc.lrms);
  write_pair(f,"queue",job_desc.queue);
  write_pair(f,"localid",job_desc.localid);
  f << "args=";
  if(job_desc.arguments.size()) {
    for(list<string>::iterator i=job_desc.arguments.begin(); \
        i!=job_desc.arguments.end(); ++i) {
      output_escaped_string(f,*i);
      f << " ";
    };
  };
  f << endl;
  write_pair(f,"subject",job_desc.DN);
  write_pair(f,"starttime",job_desc.starttime);
  write_pair(f,"lifetime",job_desc.lifetime);
  write_pair(f,"notify",job_desc.notify);
  write_pair(f,"processtime",job_desc.processtime);
  write_pair(f,"exectime",job_desc.exectime);
  write_pair(f,"rerun",inttostring(job_desc.reruns));
  if(job_desc.downloads>=0) write_pair(f,"downloads",inttostring(job_desc.downloads));
  if(job_desc.uploads>=0) write_pair(f,"uploads",inttostring(job_desc.uploads));
  write_pair(f,"jobname",job_desc.jobname);
  write_pair(f,"gmlog",job_desc.stdlog);
  write_pair(f,"cleanuptime",job_desc.cleanuptime);
  write_pair(f,"delegexpiretime",job_desc.expiretime);
  write_pair(f,"clientname",job_desc.clientname);
  write_pair(f,"clientsoftware",job_desc.clientsoftware);
  write_pair(f,"sessiondir",job_desc.sessiondir);
  write_pair(f,"diskspace",inttostring(job_desc.diskspace));
  write_pair(f,"failedstate",job_desc.failedstate);
  f.close();
  return true;
}

bool job_local_read_file(const JobId &id,const JobUser &user,JobLocalDescription &job_desc) {
  string fname = user.ControlDir() + "/job." + id + sfx_local;
  return job_local_read_file(fname,job_desc);
}

bool job_local_read_file(const string &fname,JobLocalDescription &job_desc) {
  char buf[4096];
  ifstream f(fname.c_str());
  if(! f.is_open() ) return false; /* can't open file */
  string name;
  for(;!f.eof();) {
    istream_readline(f,buf,sizeof(buf));
    name.erase();
    int p=input_escaped_string(buf,name,'=');
    if(name.length() == 0) continue;
    if(buf[p] == 0) continue;
    if(name == "lrms") { job_desc.lrms = buf+p; }
    else if(name == "queue") { job_desc.queue = buf+p; }
    else if(name == "localid") { job_desc.localid = buf+p; }
    else if(name == "subject") { job_desc.DN = buf+p; }
    else if(name == "starttime") { job_desc.starttime = buf+p; }
//    else if(name == "UI") { job_desc.UI = buf+p; }
    else if(name == "lifetime") { job_desc.lifetime = buf+p; }
    else if(name == "notify") { job_desc.notify = buf+p; }
    else if(name == "processtime") { job_desc.processtime = buf+p; }
    else if(name == "exectime") { job_desc.exectime = buf+p; }
    else if(name == "jobreport") { job_desc.jobreport = buf+p; }
    else if(name == "jobname") { job_desc.jobname = buf+p; }
    else if(name == "gmlog") { job_desc.stdlog = buf+p; }
    else if(name == "rerun") {
      string temp_s(buf+p); int n;
      if(!stringtoint(temp_s,n)) { f.close(); return false; };
      job_desc.reruns = n;
    }
    else if(name == "downloads") {
      string temp_s(buf+p); int n;
      if(!stringtoint(temp_s,n)) { f.close(); return false; };
      job_desc.downloads = n;
    }
    else if(name == "uploads") {
      string temp_s(buf+p); int n;
      if(!stringtoint(temp_s,n)) { f.close(); return false; };
      job_desc.uploads = n;
    }
    else if(name == "args") {
      job_desc.arguments.clear();
      for(int n=p;buf[n] != 0;) {
        string arg;
        n+=input_escaped_string(buf+n,arg);
        job_desc.arguments.push_back(arg);
      };
    }
    else if(name == "cleanuptime") { job_desc.cleanuptime = buf+p; }
    else if(name == "delegexpiretime") { job_desc.expiretime = buf+p; }
    else if(name == "clientname") { job_desc.clientname = buf+p; }
    else if(name == "clientsoftware") { job_desc.clientsoftware = buf+p; }
    else if(name == "sessiondir") { job_desc.sessiondir = buf+p; }
    else if(name == "failedstate") { job_desc.failedstate = buf+p; }
    else if(name == "diskspace") {
      string temp_s(buf+p);
      unsigned long long int n;
      if(!stringtoint(temp_s,n)) { f.close(); return false; };
      job_desc.diskspace = n;
    }
  };
  f.close();
  return true;
};

bool job_local_read_var(const string &fname,const string &vnam,string &value) {
  char buf[1024];
  ifstream f(fname.c_str());
  if(! f.is_open() ) return false; /* can't open file */
  string name;
  bool found = false;
  for(;!f.eof();) {
    istream_readline(f,buf,sizeof(buf));
    name.erase();
    int p=input_escaped_string(buf,name,'=');
    if(name.length() == 0) continue;
    if(buf[p] == 0) continue;
    if(name == vnam) { value = buf+p; found=true; break; };
  };
  f.close();
  return found;
};

bool job_local_read_lifetime(const JobId &id,const JobUser &user,time_t &lifetime) {
  string fname = user.ControlDir() + "/job." + id + sfx_local;
  string str;
  if(!job_local_read_var(fname,"lifetime",str)) return false;
  char* str_e;
  unsigned long int t = strtoul(str.c_str(),&str_e,10);
  if((*str_e) != 0) return false;
  lifetime=t;
  return true;
};

bool job_local_read_cleanuptime(const JobId &id,const JobUser &user,time_t &cleanuptime) {
  string fname = user.ControlDir() + "/job." + id + sfx_local;
  string str;
  if(!job_local_read_var(fname,"cleanuptime",str)) return false;
  mds_time cupt;
  cupt = str;
  cleanuptime=(time_t)cupt;
  return true;
};

bool job_local_read_notify(const JobId &id,const JobUser &user,string &notify) {
  string fname = user.ControlDir() + "/job." + id + sfx_local;
  if(!job_local_read_var(fname,"notify",notify)) return false;
  return true;
};

bool job_local_read_string(const string &fname,unsigned int num,string &str) {
  ifstream f(fname.c_str());
  if(! f.is_open() ) return false; /* can't open file */
  for(;num;num--) {
    f.ignore(INT_MAX,'\n'); 
  };
  if(f.eof()) { f.close(); return false; };
  char buf[256];
  f.get(buf,255,'\n'); if(!buf[0]) { f.close(); return false; };
  str=buf; 
  f.close();
  return true;
}

/* job.ID.input functions */

bool job_input_write_file(const JobDescription &desc,JobUser &user,list<FileData> &files) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + ".input";
  return job_Xput_write_file(fname,files) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_input_read_file(const JobId &id,JobUser &user,list<FileData> &files) {
  string fname = user.ControlDir() + "/job." + id + ".input";
  return job_Xput_read_file(fname,files);
}

/* job.ID.output functions */

bool job_cache_write_file(const JobDescription &desc,JobUser &user,list<FileData> &files) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + ".cache";
  return job_Xput_write_file(fname,files) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_output_write_file(const JobDescription &desc,JobUser &user,list<FileData> &files) {
  string fname = user.ControlDir() + "/job." + desc.get_id() + ".output";
  return job_Xput_write_file(fname,files) & fix_file_owner(fname,desc,user) & fix_file_permissions(fname);
}

bool job_cache_read_file(const JobId &id,JobUser &user,list<FileData> &files) {
  string fname = user.ControlDir() + "/job." + id + ".cache";
  return job_Xput_read_file(fname,files);
}

bool job_output_read_file(const JobId &id,JobUser &user,list<FileData> &files) {
  string fname = user.ControlDir() + "/job." + id + ".output";
  return job_Xput_read_file(fname,files);
}

/* common finctions */

bool job_Xput_write_file(const string &fname,list<FileData> &files) {
  ofstream f(fname.c_str(),ios::out | ios::trunc);
  if(! f.is_open() ) return false; /* can't open file */
  for(FileData::iterator i=files.begin();i!=files.end(); ++i) { 
    f << (*i) << endl;
  };
  f.close();
  return true;
}

bool job_Xput_read_file(list<FileData> &files) {
  for(;!cin.eof();) {
    FileData fd; cin >> fd;
//    if((fd.pfn.length() != 0) && (fd.pfn != "/")) {
    if(fd.pfn.length() != 0) {  /* returns zero length only if empty string */
      files.push_back(fd);
    };
  };
  return true;
}

bool job_Xput_read_file(const string &fname,list<FileData> &files) {
  ifstream f(fname.c_str());
  if(! f.is_open() ) return false; /* can't open file */
  for(;!f.eof();) {
    FileData fd; f >> fd;
//    if((fd.pfn.length() != 0) && (fd.pfn != "/")) {
    if(fd.pfn.length() != 0) {  /* returns zero length only if empty string */
      files.push_back(fd);
    };
  };
  f.close();
  return true;
}

bool job_clean_finished(const JobId &id,JobUser &user) {
  string fname;
  fname = user.ControlDir()+"/job."+id+".proxy"; remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+".proxy.tmp"; remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+".lrms_done"; remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+".grami"; remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+".grami_log"; remove(fname.c_str());
  return true;
}

bool job_clean_deleted(const JobId &id,JobUser &user) {
  job_clean_finished(id,user);
  string fname;
  fname = user.ControlDir()+"/job."+id+sfx_restart; remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+sfx_errors; remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+sfx_cancel; remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+sfx_clean;  remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+".output"; remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+".input"; remove(fname.c_str());
  fname = user.SessionRoot()+"/"+id+sfx_lrmsoutput; remove(fname.c_str());
  /* remove session directory */
  list<FileData> flist;
  string dname = user.SessionRoot()+"/"+id;
  delete_all_files(dname,flist,true);
  remove(dname.c_str());
  return true;
}

bool job_clean_final(const JobId &id,JobUser &user) {
  job_clean_finished(id,user);
  job_clean_deleted(id,user);
  string fname;
  fname = user.ControlDir()+"/job."+id+sfx_local;  remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+sfx_failed; remove(fname.c_str());
  job_diagnostics_mark_remove(id,user);
  job_lrmsoutput_mark_remove(id,user);
  fname = user.ControlDir()+"/job."+id+sfx_status; remove(fname.c_str());
  fname = user.ControlDir()+"/job."+id+sfx_rsl;    remove(fname.c_str());
  return true;
}

// Create multiple files for sending to logger
// TODO - make it SOAP XML so that they could be sent directly
bool job_log_make_file(const JobDescription &desc,JobUser &user,const string &url) {
  string fname_dst = user.ControlDir() + "/logs/" + desc.get_id() + ".XXXXXX";
  string fname_src;
  int h_dst,h_src;
  int l;
  bool fail = false;
  char buf[256];
  if((h_dst=mkstemp((char*)(fname_dst.c_str()))) == -1) {
    return false;
  };
  chmod(fname_dst.c_str(),S_IRUSR | S_IWUSR);
  fix_file_owner(fname_dst,desc,user);
  fix_file_permissions(fname_dst,false);
#if defined __GNUC__ && __GNUC__ >= 3
  ofstream o_dst(fname_dst.c_str());
  close(h_dst);
#else
  ofstream o_dst(h_dst);
#endif
  // URL to send info to
  if(url.length()) {
    o_dst<<"loggerurl="<<url<<endl;
    if(o_dst.fail()) { o_dst.close(); unlink(fname_dst.c_str()); return false; };
  };
  // RSL
  fname_src = user.ControlDir() + "/job." + desc.get_id() + sfx_rsl;
  h_src=open(fname_src.c_str(),O_RDONLY);
  if(h_src==-1) { o_dst.close(); unlink(fname_dst.c_str()); return false; };
  o_dst<<"rsl=";
  for(;;) {
    l=read(h_src,buf,sizeof(buf));
    if(l==0) break;
    if(l==-1) { fail=true; break; };
    for(char* p=buf;p;) { p=(char*)memchr(buf,'\r',l); if(p) (*p)=' '; };
    for(char* p=buf;p;) { p=(char*)memchr(buf,'\n',l); if(p) (*p)=' '; };
    o_dst.write(buf,l);
    if(o_dst.fail()) { fail=true; break; };
  };
  close(h_src);
  o_dst<<endl;
  // Start time
  time_t t = job_mark_time(fname_src);
  o_dst<<"starttime="<<mds_time(t)<<endl; if(o_dst.fail()) fail=true;
  // Identifier
  o_dst<<"gridid="<<desc.get_id()<<endl; if(o_dst.fail()) fail=true;
  if(fail) { o_dst.close(); unlink(fname_dst.c_str()); return false; };
  // Information from job.ID.local
  fname_src = user.ControlDir() + "/job." + desc.get_id() + sfx_local;
  h_src=open(fname_src.c_str(),O_RDONLY);
  if(h_src==-1) { o_dst.close(); unlink(fname_dst.c_str()); return false; };
  for(;;) {
    l=read(h_src,buf,sizeof(buf));
    if(l==0) break;
    if(l==-1) { fail=true; break; };
    o_dst.write(buf,l);
    if(o_dst.fail()) { fail=true; break; };
  };
  close(h_src);
  if(fail) { o_dst.close(); unlink(fname_dst.c_str()); return false; };
  // End time and failure reason
  if(desc.get_state() == JOB_STATE_FINISHED) {
    time_t t = job_state_time(desc.get_id(),user);
    if(t == 0) t=::time(NULL);
    o_dst<<"endtime="<<mds_time(t)<<endl; if(o_dst.fail()) fail=true;
    if(job_failed_mark_check(desc.get_id(),user)) {
      string failure = job_failed_mark_read(desc.get_id(),user);
      o_dst<<"failure="<<failure<<endl; if(o_dst.fail()) fail=true;
    };
    // WallTime-information from job.ID.diag
    fname_src = user.ControlDir() + "/job." + desc.get_id() + sfx_diag;
    ifstream f(fname_src.c_str());
    if(!f.is_open() ) fail=true;

    char buf[1024];
    bool continueread = true;
    while ( continueread==true && !f.eof() ) {
      istream_readline(f,buf,sizeof(buf));
      string line(buf);
      int wtstart = line.find("WallTime=");
      if( wtstart != string::npos) {
	wtstart += 9;
	int wtend = line.find("s",wtstart);
	string walltime = line.substr(wtstart,wtend-wtstart);
        int wallt = (int)atof(walltime.c_str());
	o_dst<<"walltime="<<wallt<<endl;
	continueread=false;
      }
    }
    if( continueread==true ) fail=true;
  }
  if(fail) { o_dst.close(); unlink(fname_dst.c_str()); return false; };
  o_dst.close();
  return true;
}

