#include "../std.h"

#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#include "fileroot.h"
#include "conf.h"
#include "names.h"
#include "misc.h"
#include "../misc/canonical_dir.h"
#include "../misc/log_time.h"


int FileRoot::config(Daemon &daemon,ServerParams* params) {
  /* open and read configuration file */
  ifstream cfile;
  ConfigSections* cf = NULL;
  config_open_gridftp(cfile);
  if(!cfile.is_open()) {
    olog << "Error: configuration file not found" << endl;
    return 1;
  };
  if(central_configuration) {
    cf=new ConfigSections(cfile);
    cf->AddSection("common");
    cf->AddSection("gridftpd");
  };
  for(;;) {
    string rest;
    string command;
    if(central_configuration) {
      cf->ReadNext(command,rest);
    } else {
      command = config_read_line(cfile,rest);
    };
    if(command.length() == 0) break; /* EOF */
    if(central_configuration) if(cf->SubSection()[0] != 0) continue;
    int r = daemon.config(command,rest);
    if(r == 0) continue;
    if(r == -1) { config_close(cfile); if(cf) delete cf; return 1; }
    if(command == "port") {
      if(params) {
        if(sscanf(rest.c_str(),"%u",&(params->port)) != 1) {
          olog<<"Wrong port number in configuration\n";
          config_close(cfile);
          if(cf) delete cf;
          return 1;
        };
      };
    } else if(command == "maxconnections") {
      if(params) {
        int i;
        if(sscanf(rest.c_str(),"%u",&(params->max_connections)) != 1) {
          olog<<"Wrong maxconnections number in configuration\n";
          config_close(cfile);
          if(cf) delete cf;
          return 1;
        };
      };
    } else if(command == "defaultbuffer") {
      if(params) {
        int i;
        if(sscanf(rest.c_str(),"%u",&(params->default_buffer)) != 1) {
          olog<<"Wrong defaultbuffer number in configuration\n";
          config_close(cfile);
          if(cf) delete cf;
          return 1;
        };
      };
    } else if(command == "maxbuffer") {
      if(params) {
        int i;
        if(sscanf(rest.c_str(),"%u",&(params->max_buffer)) != 1) {
          olog<<"Wrong maxbuffer number in configuration\n";
          config_close(cfile);
          if(cf) delete cf;
          return 1;
        };
      };
    } else if(command == "firewall") {
      if(params) {
        struct hostent* host;
        struct hostent  hostbuf;
        char   buf[BUFSIZ];
        int    errcode;
        string value=config_next_arg(rest);
        if(gethostbyname_r(value.c_str(),
                &hostbuf,buf,sizeof(buf),&host,&errcode)) {
          olog<<"Can't resolve host "<<value<<"\n";
          config_close(cfile);
          if(cf) delete cf;
          return 1;
        };
        if( (host->h_length < sizeof(struct in_addr)) ||
            (host->h_addr_list[0] == NULL) ) {
          olog<<"Can't resolve host "<<value<<"\n";
          config_close(cfile);
          if(cf) delete cf;
          return 1;
        };
        unsigned char* addr = (unsigned char*)(&(((struct in_addr*)(host->h_addr_list[0]))->s_addr));
        params->firewall[0]=addr[0];
        params->firewall[1]=addr[1];
        params->firewall[2]=addr[2];
        params->firewall[3]=addr[3];
      };
    } else if(!central_configuration && (command == "plugin")) {
      for(;;) {
        command=config_read_line(cfile,rest);
        if(command.length() == 0) break; /* eof - bad */
        if(command == "end") break;
      };
      continue;
    };
  };
  config_close(cfile);
  if(cf) delete cf;
  return 0;
}



int FileRoot::config(ifstream &cfile,string &pluginpath) {
  bool right_group = true;
  for(;;) {
    string rest;
    string command=config_read_line(cfile,rest);
    if(command.length() == 0) break; /* EOF */
    if(Daemon::skip_config(command) == 0) {
    }
    else if(command == "include") {  /* include content of another file */
      string name=config_next_arg(rest);
      ifstream cfile_;
      cfile_.open(name.c_str(),ifstream::in);
      if(!cfile_.is_open()) {
        olog << "Error: couldn't open file " << name << endl;
        return 1;
      };
      config(cfile_,pluginpath);
      cfile_.close();
    }
    else if(command == "encryption") {  /* is encryption allowed ? */
      string value=config_next_arg(rest);
      if(value == "yes") {
        heavy_encryption=true;
      } else if(value == "no") {
        heavy_encryption=false;
      } else {
        user.user.clear_groups();
        nodes.clear();
        olog<<"Error: improper attribute for encryption command: "<<value<<endl;
        return 1;
      };
    }
    else if(command == "allowunknown") { /* should user be present in grid-mapfile ? */
      string value=config_next_arg(rest);
      if(value == "yes") {
        user.gridmap=true;
      } else if(value == "no") {
        if(!user.gridmap) {
          olog << "Error: unknown (non-gridmap) user is not allowed" << endl;
          return 1;
        };
      } else {
        user.user.clear_groups();
        nodes.clear();
        olog<<"Error: improper attribute for encryption command: "<<value<<endl;        return 1;
      };
    }
    else if(command == "group") { /* definition of group of users based on DN */
      if(!right_group) {
        for(;;) {
          command=config_read_line(cfile,rest);
          if(command.length() == 0) break; /* eof - bad */
          if(command == "end") break;
        };
        continue;
      };
      string group_name=config_next_arg(rest);
      int decision = AAA_NO_MATCH;
      for(;;) {
        string rest=config_read_line(cfile);
        if(rest.length() == 0) break; /* eof - bad */
        if(rest == "end") break;
        if(decision == AAA_NO_MATCH) {
          decision = user.user.evaluate(rest.c_str());
          if(decision == AAA_FAILURE) decision=AAA_NO_MATCH;
        };
      };
      if(decision == AAA_POSITIVE_MATCH) user.user.add_group(group_name);
    }
    else if(command == "unixmap") {  /* map to local unix user */
      if(!unix_mapped) {
        string unix_name=config_next_arg(rest);
        string unix_group="";
        int n = unix_name.find(':');
        if(n != string::npos) {
          unix_group=unix_name.substr(n+1,string::npos);
          unix_name.erase(n,string::npos);
        };
        int decision = user.user.evaluate(rest.c_str());
        if(decision == AAA_POSITIVE_MATCH) {
          if(!user.unix_user(unix_name.c_str(),unix_group.c_str())) {
            olog<<"Wrong user specified: "<<unix_name<<" "<<unix_group<<endl;
            return 1;
          };
          unix_mapped=true;
        };
      };
    }
    else if(command == "groupcfg") {  /* next commands only for these groups */
      if(rest.find_first_not_of(" \t") == string::npos) {
        right_group=true; continue;
      };
      right_group=false;
      for(;;) {
        string group_name=config_next_arg(rest);
        if(group_name.length() == 0) break;
        right_group=user.user.check_group(group_name);
        if(right_group) break;
      };
    }
    else if(command == "pluginpath") {
      if(!right_group) continue;
      pluginpath=config_next_arg(rest);
      if(pluginpath.length() == 0) pluginpath="/";
    }
    else if(command == "port") {

    }
    else if(command == "plugin") {
      if(!right_group) {
        for(;;) {
          command=config_read_line(cfile,rest);
          if(command.length() == 0) break; /* eof - bad */
          if(command == "end") break;
        };
        continue;
      };
      string dir = config_next_arg(rest);
      string plugin = config_next_arg(rest);
      if(plugin.length() == 0) {
        olog<<"Warning: can't parse configuration line: "<<command<<" "<<dir<<" "<<plugin<<" "<<rest<< endl;
        continue;
      };
      dir=subst_user_spec(dir,&user);
      if(canonical_dir(dir,false) != 0) {
        olog<<"Warning: bad directory in plugin command: "<<dir<<endl;
        continue;
      };
      /* look if path is not already registered */
      bool already_have_this_path=false;
      for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
        if((*i) == dir) {
          already_have_this_path=true;
          break;
        };
      };
      if(already_have_this_path) {
        olog << "Already have directory:"<<dir<<endl;
        for(;;) {
          command=config_read_line(cfile,rest);
          if(command.length() == 0) break; /* eof - bad */
          if(command == "end") break;
        };
      }
      else {
        olog<<"Registering directory:"<<dir<<" with plugin: "<<plugin<<endl;
        plugin=pluginpath+'/'+plugin;
        FileNode node((char*)(dir.c_str()),(char*)(plugin.c_str)(),cfile,user);
        if(node.has_plugin()) { nodes.push_back(node); }
        else {
          olog << "Error: file node creation failed: " << dir << endl;
          for(;;) {
            command=config_read_line(cfile,rest);
            if(command.length() == 0) break; /* eof - bad */
            if(command == "end") break;
          };
        };
      };
    }
    else {
      olog << "Warning: unsupported configuration command: " << command << endl;
    };
  };
  return 0;
}

int FileRoot::config(ConfigSections &cf,string &pluginpath) {
  typedef enum {
    conf_state_single,
    conf_state_group,
    conf_state_plugin
  } config_state_t;
  config_state_t st = conf_state_single;
  bool right_group = true;
  string group_name; // =config_next_arg(rest);
  int group_decision = AAA_NO_MATCH;
  string plugin_config;
  string plugin_name;
  string plugin_path;
  for(;;) {
    string rest;
    string command;
    cf.ReadNext(command,rest);
    if(!right_group) {
      if(!cf.SectionNew()) continue;
    };
    if(cf.SectionNew()) {
      if(right_group) switch(st) {
        case conf_state_group: {
          if(group_name.length() == 0) {
            olog<<"Error: unnamed group"<<endl;
            return 1;
          };
          if(group_decision == AAA_POSITIVE_MATCH) {
            user.user.add_group(group_name);
          };
        }; break;
        case conf_state_plugin: {
          if(plugin_name.length() == 0) {
            olog<<"Warning: undefined plugin"<< endl;
            break;
          };
          if(plugin_path.length() == 0) {
            olog<<"Warning: undefined virtual plugin path"<< endl;
            break;
          };
          plugin_path=subst_user_spec(plugin_path,&user);
          if(canonical_dir(plugin_path,false) != 0) {
            olog<<"Warning: bad directory for plugin: "<<plugin_path<<endl;
            break;
          };
          /* look if path is not already registered */
          bool already_have_this_path=false;
          for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
            if((*i) == plugin_path) {
              already_have_this_path=true;
              break;
            };
          };
          if(already_have_this_path) {
            olog << "Already have directory:"<<plugin_path<<endl;
            break;
          };
          olog<<"Registering directory:"<<plugin_path<<" with plugin: "<<plugin_name<<endl;
          plugin_name=pluginpath+'/'+plugin_name;
          plugin_config+="end\n";
#ifdef HAVE_SSTREAM
          stringstream fake_cfile(plugin_config);
#else
          strstream fake_cfile;
          fake_cfile<<plugin_config;
#endif
          FileNode node((char*)(plugin_path.c_str()),(char*)(plugin_name.c_str)(),fake_cfile,user);
          if(node.has_plugin()) { nodes.push_back(node); }
          else {
            olog<<"Error: file node creation failed: "<<plugin_path<<endl;
          };
        }; break;
      };
      // prepare for new section
      plugin_config="";
      plugin_name="";
      plugin_path="";
      group_name="";
      right_group = true;
      group_decision = AAA_NO_MATCH;
      // 0 - common, 1 - group, 2 - gridftpd
      if(cf.SectionNum() == 1) {
        st=conf_state_group;
        group_name=cf.SubSection();
      } else if(cf.SubSection()[0] == 0) {
        st=conf_state_single;
      } else {
        st=conf_state_plugin;
      };
    };
    if((command.length() == 0) && (rest.length() == 0)) break; /* EOF */
    switch(st) {
      case conf_state_single: {
        if(Daemon::skip_config(command) == 0) {
        } else if(command == "include") {  /* include content of another file */
          string name=config_next_arg(rest);
          ifstream cfile_;
          cfile_.open(name.c_str(),ifstream::in);
          if(!cfile_.is_open()) {
            olog << "Error: couldn't open file " << name << endl;
            return 1;
          };
          config(cfile_,pluginpath); // included are in old format
          cfile_.close();
        } else if(command == "encryption") {  /* is encryption allowed ? */
          string value=config_next_arg(rest);
          if(value == "yes") {
            heavy_encryption=true;
          } else if(value == "no") {
            heavy_encryption=false;
          } else {
            user.user.clear_groups();
            nodes.clear();
            olog<<"Error: improper attribute for encryption command: "<<value<<endl;
            return 1;
          };
        } else if(command == "allowunknown") {
          /* should user be present in grid-mapfile ? */
          string value=config_next_arg(rest);
          if(value == "yes") {
            user.gridmap=true;
          } else if(value == "no") {
            if(!user.gridmap) {
              olog<<"Error: unknown (non-gridmap) user is not allowed"<<endl;
              return 1;
            };
          } else {
            user.user.clear_groups();
            nodes.clear();
            olog<<"Error: improper attribute for encryption command: "<<value<<endl;
            return 1;
          };
        } else if(command == "pluginpath") {
          if(!right_group) continue;
          pluginpath=config_next_arg(rest);
          if(pluginpath.length() == 0) pluginpath="/";
        } else if(command == "port") {
        } else if(command == "unixmap") {  /* map to local unix user */
          if(!unix_mapped) {
            string unix_name=config_next_arg(rest);
            string unix_group="";
            int n = unix_name.find(':');
            if(n != string::npos) {
              unix_group=unix_name.substr(n+1,string::npos);
              unix_name.erase(n,string::npos);
            };
            int decision = user.user.evaluate(rest.c_str());
            if(decision == AAA_POSITIVE_MATCH) {
              if(!user.unix_user(unix_name.c_str(),unix_group.c_str())) {
               olog<<"Wrong user specified: "<<unix_name<<" "<<unix_group<<endl;
               return 1;
              };
              unix_mapped=true;
            };
          };
        };
      }; break;
      case conf_state_group: {
        /* definition of group of users based on DN */
        if(command == "name") {
          group_name=rest;
        } else if(command == "groupcfg") {
          if(rest.find_first_not_of(" \t") == string::npos) {
            right_group=true;
          } else {
            right_group=false;
            for(;;) {
              string group_name=config_next_arg(rest);
              if(group_name.length() == 0) break;
              right_group=user.user.check_group(group_name);
              if(right_group) break;
            };
          };
        } else {
          if(group_decision == AAA_NO_MATCH) {
            rest=command+" "+rest;
            group_decision = user.user.evaluate(rest.c_str());
            if(group_decision == AAA_FAILURE) group_decision=AAA_NO_MATCH;
          };
        };
      }; break;
      case conf_state_plugin: {
        if(command == "groupcfg") {
          if(rest.find_first_not_of(" \t") == string::npos) {
            right_group=true;
          } else {
            right_group=false;
            for(;;) {
              string group_name=config_next_arg(rest);
              if(group_name.length() == 0) break;
              right_group=user.user.check_group(group_name);
              if(right_group) break;
            };
          };
        } else if(command == "plugin") {
          plugin_name=rest;
        } else if(command == "path") {
          plugin_path=rest;
        } else {
          plugin_config+=command;
          plugin_config+=" ";
          plugin_config+=rest;
          plugin_config+="\n";
        };
      }; break;
    };
  };
  return 0;
}

int FileRoot::config(globus_ftp_control_auth_info_t *auth,
                     gss_cred_id_t delegated_cred,
                     globus_ftp_control_handle_t *handle) {
  /* open and read configuration file */
  ifstream cfile;
  ConfigSections* cf = NULL;
  config_open_gridftp(cfile);
  if(!cfile.is_open()) {
    olog << "Error: configuration file not found" << endl;
    return 1;
  };
  if(central_configuration) {
    cf=new ConfigSections(cfile);
    cf->AddSection("common");
    cf->AddSection("group");
    cf->AddSection("gridftpd");
  };
  /* keep information about user */
  if(fill_user_spec(&user,auth,delegated_cred,handle) != 0) return 1;
  string pluginpath="/";
  int r;
  if(!central_configuration) {
    r = config(cfile,pluginpath);
  } else {
    r = config(*cf,pluginpath);
  };
  config_close(cfile);
  if(cf) delete cf;
  if(r != 0) return r;
  if(!user.gridmap) {
    olog << "Error: unknown (non-gridmap) user is not allowed" << endl;
    return 1;
  };
  /* must be sorted to make sure we can always find right directory */
  nodes.sort(FileNode::compare);
  /* create dummy directories */
  int nn=nodes.size();
  for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
    if(nn==0) break;
    string name = i->point;
    for(;remove_last_name(name);) {
      if(name.length() == 0) break;
      bool found = false;
      list<FileNode>::iterator ii=i;
      for(;;) {
        ++ii; if(ii == nodes.end()) break;
        if(name == ii->point) { found=true; break; };
      };
      if(!found) { /* add dummy dir */
        olog<<"Registering dummy directory:"<<name<<endl;
        nodes.push_back(FileNode(name.c_str()));
      }
      else {
        break;
      };
    };
    nn--;
  };
  opened_node=nodes.end();
  user.free();
  return 0;
}

