#include "../std.h"

#include <string>

#include <globus_gss_assist.h>
#include <globus_io.h>
#include <globus_ftp_control.h>
#include "userspec.h"
#include "../misc/escaped.h"
#include "../misc/log_time.h"
#include "../misc/inttostring.h"
#include "conf.h"

void userspec_t::free(void) {
//  free(spec->subject);
  if(home)  std::free(home);  home=NULL;
  if(name)  std::free(name);  name=NULL;
  if(group) std::free(group); group=NULL;
//  Groups.clear();
  // Keep authentication info to preserve proxy (just in case)
}

userspec_t::userspec_t(void):name(NULL),group(NULL),home(NULL),gridmap(false) {
}

userspec_t::~userspec_t(void) {
  userspec_t::free();
}

bool userspec_t::unix_user(const char* name_,const char* group_) {
  if(getuid() != 0) return true; // senseless

  struct passwd pw_;
  struct group gr_;
  struct passwd *pw;
  struct group *gr;
  char buf[BUFSIZ];
  char* name_tmp = NULL;
  char* home_tmp = NULL;

  getpwnam_r(name_,&pw_,buf,BUFSIZ,&pw);
  if(pw == NULL) {
    olog<<"Local user "<<name_<<" does not exist"<<endl;
    return false;
  };
  name_tmp=strdup(pw->pw_name);
  home_tmp=strdup(pw->pw_dir);
  if(group_ && group_[0]) {
    getgrnam_r(group_,&gr_,buf,BUFSIZ,&gr);
    if(gr == NULL) {
      olog<<"Local group "<<group_<<" does not exist"<<endl;
      if(name_tmp) std::free(name_tmp); 
      if(home_tmp) std::free(home_tmp);
      return false;
    };
  } else {
    getgrgid_r(pw->pw_gid,&gr_,buf,BUFSIZ,&gr);
    if(gr == NULL) {
      olog<<"User "<<name_<<" has invalid local group id "<<pw->pw_gid<<endl;
      if(name_tmp) std::free(name_tmp); 
      if(home_tmp) std::free(home_tmp);
      return false;
    };
  };
  if(name) std::free(name);
  if(group) std::free(group);
  if(home) std::free(home);
  name=name_tmp;  uid=pw->pw_uid;
  group=strdup(gr->gr_name); gid=gr->gr_gid;
  home=home_tmp;
  olog<<"Remapped to local user: "<<name<<endl;
  olog<<"Remapped to local id: "<<uid<<endl;
  olog<<"Remapped to local group id: "<<gid<<endl;
  olog<<"Remapped to local group name: "<<group<<endl;
  olog<<"Remapped user's home: "<<home<<endl;
  return true;
}

bool check_gridmap(const char* dn,char** user,const char* mapfile) {
  string globus_gridmap;
  if(mapfile) {
    globus_gridmap=mapfile;
  }
  else {
    char* tmp=getenv("GRIDMAP");
    if((tmp == NULL) || (tmp[0] == 0)) {
      globus_gridmap="/etc/grid-security/grid-mapfile";
    }
    else { globus_gridmap=tmp; };
  };
  ifstream f(globus_gridmap.c_str());
  if(!f.is_open() ) {
    olog<<"Mapfile is missing at "<<globus_gridmap<<endl;
    return false;
  };
  for(;!f.eof();) {
    char buf[512]; // must be enough for DN + name
    istream_readline(f,buf,sizeof(buf));
    buf[511]=0;
    char* p = buf;
    for(;*p;p++) if(((*p) != ' ') && ((*p) != '\t')) break;
    if((*p) == '#') continue;
    if((*p) == 0) continue;
    string val;
    int n = input_escaped_string(p,val);
    if(strcmp(val.c_str(),dn) != 0) continue;
    p+=n;
    if(user) {
      n=input_escaped_string(p,val);
      *user=strdup(val.c_str());
    };
    f.close();
    return true;
  };
  f.close();
  return false;
}

int fill_user_spec(userspec_t *spec,globus_ftp_control_auth_info_t *auth,gss_cred_id_t delegated_cred,globus_ftp_control_handle_t *handle) {
  struct passwd pw_;
  struct group gr_;
  struct passwd *pw;
  struct group *gr;
  char buf[BUFSIZ];

  if(auth->auth_gssapi_subject == NULL) return 1;
  string subject = auth->auth_gssapi_subject;
  make_unescaped_string(subject);
  spec->name=NULL;
  if(!check_gridmap(subject.c_str(),&(spec->name))) {
    olog<<"Warning: there is no local mapping for user"<<endl;
  } else {
    if((spec->name == NULL) || (spec->name[0] == 0)) {
      olog<<"Warning: there is no local name for user"<<endl;
    } else {
      spec->gridmap=true;
    };
  };
  spec->user.set(auth->auth_gssapi_subject,delegated_cred);
  if((spec->user.proxy() == NULL) || (spec->user.proxy()[0] == 0)) {
    olog<<"No proxy provided"<<endl;
  } else {
    olog<<"Proxy stored at "<<spec->user.proxy()<<endl;
  };
  if((getuid() == 0) && spec->name && spec->name[0]) {
    olog<<"Mapped to local user: "<<spec->name<<endl;
    getpwnam_r(spec->name,&pw_,buf,BUFSIZ,&pw);
    if(pw == NULL) {
      olog<<"Local user does not exist"<<endl;
//      free(spec->subject); spec->subject=NULL;
      free(spec->name); spec->name=NULL;
      return 1;
    };
  }
  else {
    if(spec->name) free(spec->name); spec->name=NULL;
    getpwuid_r(getuid(),&pw_,buf,BUFSIZ,&pw);
    if(pw == NULL) {
      olog<<"Warning: running user has no name"<<endl;
//      spec->subject=strdup("unknown");
    }
    else {
      spec->name=strdup(pw->pw_name);
      olog<<"Mapped to running user: "<<spec->name<<endl;
    };
  };
  if(!(spec->name)) spec->name=strdup("");
  spec->uid=pw->pw_uid;
  spec->gid=pw->pw_gid;
  olog<<"Mapped to local id: "<<spec->uid<<endl;
  olog<<"Mapped to local group id: "<<spec->gid<<endl;
  spec->home=strdup(pw->pw_dir);
  olog<<"Mapped user's home: "<<spec->home<<endl;
  getgrgid_r(spec->gid,&gr_,buf,BUFSIZ,&gr);
  if(gr == NULL) {
    olog<<"Invalid local group"<<endl;
//    if(spec->subject) free(spec->subject); spec->subject=NULL;
    if(spec->name) free(spec->name); spec->name=NULL;
    if(spec->home) free(spec->home); spec->home=NULL;
    return 1;
  };
  spec->group=strdup(gr->gr_name);
  olog<<"Mapped to local group name: "<<spec->group<<endl;
//  spec->delegated_cred=delegated_cred;
/* fill host info here */
  if(globus_io_tcp_get_remote_address(&(handle->cc_handle.io_handle),
         spec->host,&(spec->port)) != GLOBUS_SUCCESS) {
    spec->port=0;
  };
  return 0;
}

string subst_user_spec(string &in,userspec_t *spec) {
  string out = "";
  int i;
  int last;
  last=0;
  for(i=0;i<in.length();i++) {
    if(in[i] == '%') {
      if(i>last) out+=in.substr(last,i-last);
      i++; if(i>=in.length()) {         };
      switch(in[i]) {
        case 'u': {
          char buf[10];
          snprintf(buf,9,"%i",spec->uid); out+=buf; last=i+1; 
        }; break;
        case 'U': { out+=spec->name; last=i+1; }; break;
        case 'g': {
          char buf[10];
          snprintf(buf,9,"%i",spec->gid); out+=buf; last=i+1; 
        }; break;
        case 'G': { out+=spec->group; last=i+1; }; break;
        case 'D': { out+=spec->user.DN(); last=i+1; }; break;
        case 'H': { out+=spec->home; last=i+1; }; break;
        case '%': { out+='%'; last=i+1; }; break;
        default: {
          olog << "Warning: undefined control sequence: %" << in[i] << endl;
        };
      };
    };
  };
  if(i>last) out+=in.substr(last);
  return out;
}
