// Request data transfer (and optionally wait till completed).

#include "std.h"
#include "datamove/datapoint.h"
#include "datamove/datahandle.h"
#include "misc/log_time.h"
#include "misc/globus_error_utils.h"
#include "misc/condition.h"
#include "../userinterface/CertInfo.h"

// SSE
#include "https/se/file_soapH.h"
extern SOAP_NMAC struct Namespace file_soap_namespaces[];
#include "https/client/client.h"


bool sse_replicate(DataPoint& dest,list<string>& sources,bool verbose);
bool ftp_replicate(DataPoint& dest,list<string>& sources,bool verbose);
bool meta_replicate(DataPoint& dest,list<string>& sources,bool verbose);

bool ftp_ftp_replicate(const char* dest,const char* src,bool verbose);

int main(int argc, char* argv[]) {
  LogTime::active=false;
  bool verbose = false;
  list<string> sources;
  for(;;) {
    opterr=0;
    int optc=getopt(argc,argv,"+hvd:s:");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"ngrequest [-h] [-v] [-d level] [-s source] destination"<<endl;
        olog<<" -h - print help"<<endl;
        olog<<" -v - print version"<<endl;
        olog<<" -d - set debug level"<<endl;
        olog<<" -s - source of data (can be multiple)"<<endl;
        return 1;
      }; break;
      case 'v': {
        olog<<"ngls: version "<<VERSION<<endl;
        return 0;
      };
      case 's': {
        sources.push_back(string(optarg));
      }; break;
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<endl;
          return 1;
        };
        verbose=true;
        LogTime::level=i;
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        return 1;
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        return 1;
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        return 1;
      };
    };
  };
  char* service_url = argv[optind+0];
  if(service_url == NULL) {
    olog<<"Missing service url" << endl; return 1;
  };
  if(argv[optind+1] != NULL) {
    olog<<"Unrecognized option: "<<argv[optind+1]<<endl; return 1;
  };
  DataPoint url(service_url);
  if(!url) {
    olog<<"Unsupported service url given"<<endl; return 1;
  };
  if(url.meta()) {
    if(!meta_replicate(url,sources,verbose)) {
      olog<<"Failed"<<endl; return 1;
    };
  } else if(strncasecmp(service_url,"se://",5) == 0) {
    // send request to SSE 
    if(!sse_replicate(url,sources,verbose)) {
      olog<<"Failed"<<endl; return 1;
    };
  } else if(strncasecmp(service_url,"gsiftp://",9) == 0) {
    if(!ftp_replicate(url,sources,verbose)) {
      olog<<"Failed"<<endl; return 1;
    };
    //olog<<"GridFTP is not supported yet"<<endl; return 1;
  } else {
    olog<<"URL "<<service_url<<" is not supported"<<endl; return 1;
  };
  return 0;
}

// Destination is specified as meta-URL. Need to find proper destinations
// and initiate transfers.
bool meta_replicate(DataPoint& dest,list<string>& sources,bool verbose) {
  if(!dest.meta_resolve(false,UrlMap())) {
    odlog(0)<<"Failed to resolve possible destinations"<<endl;
    return false;
  };
  if(!dest.have_locations()) {
    odlog(0)<<"No destination locations found"<<endl;
    return false;
  };
  bool replication = (sources.size() == 0);
  for(;dest.have_location();dest.next_location()) {
    const string& loc = dest.current_location();
    if(strncasecmp(loc.c_str(),"se://",5) == 0) {
      odlog(1)<<"Destination chosen: "<<loc<<endl;
      if(!dest.meta_preregister(replication,false)) {
        odlog(1)<<"Failed to pre-register"<<endl; continue;
      };
      if(sse_replicate(dest,sources,verbose)) {
        if(!dest.meta_postregister(replication,false)) {
          odlog(1)<<"Failed to post-register"<<endl;
          dest.meta_preunregister(replication);
          continue;
        };
        return true;
      };
      if(!dest.meta_preunregister(replication)) {
        odlog(1)<<"WARNING: Failed to un-register failed transfer."<<endl;
      };
      odlog(1)<<"Failed - trying next destination"<<endl;
    } else if((strncasecmp(loc.c_str(),"gsiftp://",9) == 0) ||
              (strncasecmp(loc.c_str(),"ftp://",6) == 0)) {
      odlog(1)<<"Destination chosen: "<<loc<<endl;
      if(!dest.meta_preregister(replication,false)) {
        odlog(1)<<"Failed to pre-register"<<endl; continue;
      };
      if(ftp_replicate(dest,sources,verbose)) {
        if(!dest.meta_postregister(replication,false)) {
          odlog(1)<<"Failed to post-register"<<endl;
          dest.meta_preunregister(replication);
          continue;
        };
        return true;
      };
      if(!dest.meta_preunregister(replication)) {
        odlog(1)<<"WARNING: Failed to un-register failed transfer."<<endl;
      };
    } else {
      odlog(1)<<"Unsupported location "<<loc<<endl;
    };
  };
}

bool sse_replicate(DataPoint& dest,list<string>& sources,bool verbose) {
  string service_url = dest.current_location();
  service_url.replace(0,2,"httpg");
  string service_path = "";
  {
    int n = service_url.find('?');
    if(n == string::npos) {
      odlog(0)<<"Missing LFN in destination URL"<<endl;
      return false;
    };
    service_path=service_url.c_str()+n+1;
    service_url.resize(n);
  };
  canonic_url(service_url); // remove options
  odlog(1)<<"Talking to SOAP service at "<<service_url<<endl;
  struct soap soap;
  HTTP_ClientSOAP s(service_url.c_str(),&soap);
  soap.namespaces=file_soap_namespaces;
  if(s.connect() != 0) {
    odlog(0)<<"Failed to connect to "<<service_url<<endl;
    return false;
  };
  ns__fileinfo info;
  ns__addResponse rr;
  int soap_err = SOAP_OK;
  info.id=(char*)(service_path.c_str());
  string acl("");
  {
    CertInfo ci;
    acl=ci.GetSN();
    if(acl.length()) acl="<gacl><entry><person><dn>"+acl+"</dn></person><allow><read/><write/><list/><admin/></allow></entry></gacl>";
    info.acl=(char*)(acl.c_str());
  };
  // unsigned long long int* size;
  // char* checksum;
  // char* created;
  // enum ns__filestate* state;
  // int __size_url;
  // char** url;
  int size_sources_ = sources.size();
  char** sources_;
  if(size_sources_ == 0) {
    sources_=(char**)malloc(sizeof(char*));
    sources_[0]=""; // replication
    size_sources_=1;
  } else {
    sources_=(char**)malloc(sizeof(char*)*size_sources_);
    int i = 0;
    list<string>::iterator ii = sources.begin();
    for(;i<size_sources_;++i,++ii) sources_[i]=(char*)(ii->c_str());
  };
  // Send resuest to server for new file
  if((soap_err=soap_call_ns__add(&soap,s.SOAP_URL(),"add",&info,
                           size_sources_,sources_,rr)) != SOAP_OK) {
    odlog(1)<<"Failed to execute remote soap call 'add' at "<<service_url<<endl;
    return false;
  };
  if(rr.error_code != 0) {
    odlog(1)<<"Failed ("<<rr.error_code<<") to create remote file "<<info.id<<" at "<<service_url<<endl;
    return false;
  };
  s.reset();
  // Wait till either file disappears (failed) or reaches proper state
  string file_url=s.SOAP_URL(); file_url+="/"; file_url+=service_path;
  bool file_detected = false;
  int retries = 5;
  time_t start = time(NULL);
  for(;;) {
    ns__infoResponse rr;
    if((soap_err=soap_call_ns__info(&soap,file_url.c_str(),"info",
                           NULL,rr)) != SOAP_OK) {
      odlog(1)<<"Failed to execute remote soap call 'info' at "<<service_url<<endl;
      retries--;
      if((--retries) <= 0) {
        odlog(0)<<"Can't track state of file at "<<service_url<<endl;
        return false;
      };
      sleep(5);
      s.reset();
      continue;
    };
    retries=5;
    if((rr.error_code != 0) || (rr.__size_file == 0)) {
      odlog(1)<<"Error "<<rr.error_code<<"."<<rr.sub_error_code<<": "<<rr.error_description<<endl;
      if(file_detected) {
        odlog(0)<<"Download failed."<<endl;
      } else {
        odlog(0)<<"File disappeared before client could see it. Most probably download failed."<<endl;
      };
      return false;
    };
    if(rr.__size_file != 1) {
      odlog(0)<<"Wrong number of files reported. Must be some error at  server's side."<<endl;
      return false;
    };
    if(verbose) {
      const char* fstate="unknown     ";
      if(rr.file[0].state) {
        switch(*(rr.file[0].state)) {
          case 0: fstate="ACCEPTED    "; break;
          case 1: fstate="COLLECTING  "; break;
          case 2: fstate="REQUESTED   "; break;
          case 3: fstate="DOWNLOADING "; break;
          case 4: fstate="COMPLETE    "; break;
          case 5: fstate="VALID       "; break;
        };
      };
      odlog(0)<<"\rTime passed: "<<time(NULL)-start<<" s, remote state: "<<fstate<<flush;
    };
    if(rr.file[0].state && (*(rr.file[0].state) == 5)) {
      odlog(0)<<"File downloaded"<<endl;
      return true;
    };
    s.reset();
    sleep(5);
  };
  return true;
}

Condition<int> ftp_cond;

void ftp_replicate_callback(void *user_arg,globus_ftp_client_handle_t *handle,globus_object_t *error) {
  if(error != GLOBUS_SUCCESS) {
    cerr<<"ftp_replicate_callback failed: "<<error<<endl;
    ftp_cond.signal(1);
  } else {
    ftp_cond.signal(0);
  };
}

bool ftp_replicate(DataPoint& dest,list<string>& sources,bool verbose) {
  if(sources.size() == 0) {
    if(!dest.meta()) return false;
    list<string> sources_;
    sources_.push_back(dest.base_url());
    return ftp_replicate(dest,sources_,verbose);
  };
  list<DataPoint*> srcs;
  for(list<string>::iterator s = sources.begin();s!=sources.end();++s) {
    DataPoint* url = new DataPoint(s->c_str());
    if(*url) {
      srcs.push_back(url);
    } else {
      odlog(0)<<"Unsupported URL: "<<*s<<endl;
      delete url;
    };
  };
  for(;;) {
    bool transfer_tried = false;
    for(list<DataPoint*>::iterator src = srcs.begin();src!=srcs.end();++src) {
      DataPoint& source = **src;
      if(!source.meta_resolve(true,UrlMap())) {
        odlog(0)<<"Failed to resolve source: "<<source<<endl;
        source.next_location(); /* try again */
        continue;
      };
      if(!source.have_location()) continue;
      if(source.current_location() == dest.current_location()) continue;
      if((strncasecmp(source.current_location(),"ftp://",6) == 0) ||
         (strncasecmp(source.current_location(),"gsiftp://",9) == 0)) {
        odlog(1)<<"Source chosen: "<<source.current_location()<<endl;
        transfer_tried=true;
        if(ftp_ftp_replicate(dest.current_location(),source.current_location(),verbose)) {
          for(list<DataPoint*>::iterator s=srcs.begin();s!=srcs.end();++s) delete *s;
          return true;
        } else {
          odlog(1)<<"Failed transfer"<<endl;
          source.next_location();
        };
      } else {
        odlog(1)<<"Can't transfer to (gsi)FTP from source "<<source.current_location()<<endl;
        source.next_location();
      };
    };
    if(!transfer_tried) break;
  };
  for(list<DataPoint*>::iterator s=srcs.begin();s!=srcs.end();++s) delete *s;
  return false;
}

bool ftp_ftp_replicate(const char* dest,const char* src,bool verbose) {
  globus_ftp_client_handle_t handle;
  globus_ftp_client_handleattr_t attr;
  globus_ftp_client_operationattr_t src_attr;
  globus_ftp_client_operationattr_t dest_attr;
  globus_result_t res;

  globus_ftp_client_handleattr_init(&attr);
  globus_ftp_client_handle_init(&handle,&attr);
  globus_ftp_client_operationattr_init(&src_attr);
  globus_ftp_client_operationattr_init(&dest_attr);


  globus_module_activate(GLOBUS_FTP_CLIENT_MODULE);

  res=globus_ftp_client_third_party_transfer(&handle,
   src,&src_attr, dest,&dest_attr,
   GLOBUS_NULL,&ftp_replicate_callback,GLOBUS_NULL);
  if(res != GLOBUS_SUCCESS) {
    cerr<<"globus_ftp_client_third_party_transfer failed: "<<GlobusResult(res)<<endl;
    globus_module_deactivate(GLOBUS_FTP_CLIENT_MODULE);
    return false;
  };
  int r;
  ftp_cond.wait(r);
  globus_module_deactivate(GLOBUS_FTP_CLIENT_MODULE);
  if(r != 0) return false;
  return true;
}
