/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*

Shared authentication and authorization module used by all clients.
Gridssh/scp/sftp clients use globus I/O with authentication and encryption 
included.

*/

#include "runtime_conf.h"
#include "gridssh.h"
#include "log.h"
#include "file_io.h"
#include "home_dir.h"
#include "auth_client.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* basename */
#include <libgen.h>

/* getopt */
#include <unistd.h>

/* fork/waitpid*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

/* globus */
#include <globus_io.h>
#include <globus_gsi_credential.h>

/*  Global constants  */
extern runtime_conf *conf;

int globus_activation(void)
{
  int err = 0;
  int rval;

  rval = globus_module_activate(GLOBUS_IO_MODULE);
  if (rval != GLOBUS_SUCCESS) {
    errmsg("Globus I/O module activation failed");
    err++;
  }
  
  rval = globus_module_activate(GLOBUS_GSI_GSS_ASSIST_MODULE);
  if(rval != GLOBUS_SUCCESS) {
    errmsg("Globus assist module activation failed");
    err++;
  }

  rval = globus_module_activate(GLOBUS_GSI_CREDENTIAL_MODULE);
  if(rval != GLOBUS_SUCCESS) {
    errmsg("Globus credential module activation failed");
    err++;
  }

  return err;
}


int globus_deactivation(void)
{
  int err = 0;
  int rval;

  rval = globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
  if(rval != GLOBUS_SUCCESS) {
    errmsg("Globus credential module deactivation failed");
    err++;
  }
  rval = globus_module_deactivate(GLOBUS_GSI_GSS_ASSIST_MODULE);
  if(rval != GLOBUS_SUCCESS) {
    errmsg("Globus assist module deactivation failed");
    err++;
  }

  rval = globus_module_deactivate(GLOBUS_IO_MODULE);
  if (rval != GLOBUS_SUCCESS) {
    errmsg("Globus I/O module deactivation failed");
    err++;
  }
  
  return err;
}

auth_status *authorize(char *server, char *port)
{
  auth_status *status = NULL;
  /* status response */
  char reply[MAX_STATUS_LEN];
  
  /* Variables for Globus functions */
  globus_result_t rptr;
  globus_object_t *err;
  globus_io_attr_t attr;
  globus_io_handle_t conn_handle;
  globus_size_t nbytes;

  /* Variables for security functionality */
  globus_io_secure_authorization_data_t auth_data;

  time_t now;
  char *status_str = NULL, *user_str = NULL, *expire_str = NULL;
  char *home_dir = NULL, *cache_filename = NULL, *cache_line = NULL;
  char *search = NULL;

  status = malloc(sizeof(auth_status));
  assert(status != NULL);
  
  status->server = malloc((strlen(server)+1)*sizeof(char));
  assert(status->server != NULL);
  strcpy(status->server, server);
  status->port = malloc((strlen(port)+1)*sizeof(char));
  assert(status->port != NULL);
  strcpy(status->port, port);
  status->ssh_user = NULL;
  status->expire = (time_t)NOT_AUTHORIZED;

  /* get time before login to make sure expire is always before actual expire */
  time(&now);

  /* Check if we need to do auth */
  if(conf->use_cache) {

    if((home_dir = get_home()) == NULL) {
      errmsg("failed to retrieve home_dir - can't continue");
      exit(EXIT_FAILURE);
    }
    
    cache_filename = malloc((strlen(home_dir)+1+strlen(conf->cache_filename)+1)*sizeof(char));
    assert(cache_filename != NULL);
    sprintf(cache_filename,"%s/%s",home_dir,conf->cache_filename);
    free(home_dir);
    
    debugmsg(4,"reading auth info from cache file %s", cache_filename);
    search = malloc((strlen(status->server)+1+strlen(status->port)+1+1)*sizeof(char));
    assert(search != NULL);
    sprintf(search, "%s;%s;", status->server,status->port);

    cache_line = match_line(search, cache_filename);
    free(search);

    if(cache_line == NULL) {
      debugmsg(3,"cache file didn't contain relevant auth info for %s", status->server);
    } else {
      debugmsg(4,"found matching line %s in cache", cache_line);
      /* must have enough space to contain entire cache_line string */
      user_str = malloc((strlen(cache_line)+1)*sizeof(char));
      assert(user_str != NULL);
      expire_str = malloc((strlen(cache_line)+1)*sizeof(char));
      assert(expire_str != NULL);
      debugmsg(6,"scanning line %s for user and expiry", cache_line);
      /* ignore first three matched patterns */
      if(sscanf(cache_line, "%*[A-Za-z0-9._-];%*[0-9];%[a-z0-9-];%[0-9]", \
		user_str, expire_str) == 2) {
	debugmsg(4,"setting ssh_user to cached value of %s", user_str);
	status->ssh_user = malloc((strlen(user_str)+1)*sizeof(char));
	assert(status->ssh_user != NULL);
	strcpy(status->ssh_user, user_str);
	
	free(user_str);
	
	debugmsg(4,"found expire string %s, now is %ld", expire_str, now);
	
	status->expire = (time_t)(atol(expire_str));
	free(expire_str);
	/* add a little slack to avoid time running out during login */
	if(now + DEFAULT_LOGIN_TIMEOUT/1000000 < status->expire) {
	  debugmsg(1,"cached auth info: proxy is still valid for %ld seconds - skipping auth", status->expire - now);
	  return status;
	} else {
	  debugmsg(1,"cached auth info: proxy expired or about to: now + timeout is %ld, expire is %ld  - doing auth", now + + DEFAULT_LOGIN_TIMEOUT/1000000, status->expire);
	  status->expire = NOT_AUTHORIZED;
	}
      } else {
	errmsg("cache line is not in expected format! tampering? ignoring!");
	free(user_str);
	free(expire_str);
      }
    }
  }
  
  /* Cache disabled or didn't contain relevant auth info - Negotiate auth */

  if(globus_activation()) {
    errmsg("Activation of required globus modules failed - can't continue");
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }
    
    /* Initialize the structure attr. */
    globus_io_tcpattr_init(&attr);
  
  rptr = globus_io_attr_set_secure_authentication_mode(&attr, 
						       GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI, GSS_C_NO_CREDENTIAL);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("Setting secure authentication mode failed");
    globus_io_tcpattr_destroy(&attr);
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }
  
  /* Initialize the authorization data */
  rptr = globus_io_secure_authorization_data_initialize(&auth_data);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("Initializing authorization data failed");
    globus_io_secure_authorization_data_destroy(&auth_data);
    globus_io_tcpattr_destroy(&attr);
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }
  
  /* Set the authorization mode to require a valid host certificate */
  rptr = globus_io_attr_set_secure_authorization_mode(&attr, 
						      GLOBUS_IO_SECURE_AUTHORIZATION_MODE_HOST, &auth_data);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("Setting authorization mode failed");
    globus_io_secure_authorization_data_destroy(&auth_data);
    globus_io_tcpattr_destroy(&attr);
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }
  
  /* We transfer the public key using full proxy delegation */
  rptr = globus_io_attr_set_secure_delegation_mode(&attr, GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("Setting delegation mode failed");
    globus_io_secure_authorization_data_destroy(&auth_data);
    globus_io_tcpattr_destroy(&attr);
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }
  
  /* Set the channel mode to use wrapping */
  rptr = globus_io_attr_set_secure_channel_mode(&attr, 
						GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("Setting channel mode failed");
    globus_io_secure_authorization_data_destroy(&auth_data);
    globus_io_tcpattr_destroy(&attr);
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }
  
  /* Use encryption, SAFE instead of PRIVATE is also possible
     but gives only integrity protection */
  rptr = globus_io_attr_set_secure_protection_mode(&attr, 
						   GLOBUS_IO_SECURE_PROTECTION_MODE_PRIVATE);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("Setting encrypted mode failed");
    globus_io_secure_authorization_data_destroy(&auth_data);
    globus_io_tcpattr_destroy(&attr);
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }   


  /*  Connect to the gridsshd server to gain access */
  rptr = globus_io_tcp_connect(status->server, (unsigned short int)atoi(status->port), &attr, &conn_handle);
  if (rptr != GLOBUS_SUCCESS) {
    /* Expand the error object and report the type of failure */
    err = globus_error_get(rptr);
    if (globus_object_type_match(GLOBUS_IO_ERROR_TYPE_AUTHENTICATION_FAILED, 
				 globus_object_get_type(err))) {
      errmsg("Authentication failed");
    }
    else if (globus_object_type_match(GLOBUS_IO_ERROR_TYPE_AUTHORIZATION_FAILED,
				      globus_object_get_type(err))) {
      errmsg("Authorization failed");
    }
    else if (globus_object_type_match(GLOBUS_IO_ERROR_TYPE_NO_CREDENTIALS,
				      globus_object_get_type(err))) {
      errmsg("Failure: no credentials - please make sure that a valid proxy is available");
    }
    else {
      errmsg("Connecting to the remote auth server failed (%x)", rptr);
    }
    
    globus_object_free(err);	
    globus_io_close(&conn_handle);
    globus_io_secure_authorization_data_destroy(&auth_data);
    globus_io_tcpattr_destroy(&attr);
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }
  
  debugmsg(3,"Initial authentication and authorization with server succeeded");
  
  /* Get status of further authorization */
  rptr = globus_io_read(&conn_handle, (globus_byte_t *)reply, MAX_STATUS_LEN-1, 1, &nbytes);
  
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("Reading server status from the socket failed");
    globus_io_close(&conn_handle);
    globus_io_secure_authorization_data_destroy(&auth_data);
    globus_io_tcpattr_destroy(&attr);
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }
  
  /*  Check status */
  reply[nbytes]='\0';
  
  debugmsg(5,"Server sent authorization status string: %s", reply);
  
  /* must have enough space to contain entire cache_line string */
  status_str = malloc((nbytes+1)*sizeof(char));
  assert(status_str != NULL);
  user_str = malloc((nbytes+1)*sizeof(char));
  assert(user_str != NULL);
  expire_str = malloc((nbytes+1)*sizeof(char));
  assert(expire_str != NULL);
  /*  Strict matching to avoid abuse */
  if((sscanf(reply, "%[A-Z];%[A-Za-z0-9];%[0-9]", \
	     status_str, user_str, expire_str) > 0) && \
     (!strcmp(status_str,SUCCESS_STR))) {
    debugmsg(3,"remaining authorization with server succeeded");
  } else {
    errmsg("remaining authorization with server failed - exitting");
    free(status_str);
    free(user_str);
    free(expire_str);
    globus_io_close(&conn_handle);
    globus_io_secure_authorization_data_destroy(&auth_data);
    globus_io_tcpattr_destroy(&attr);
    globus_deactivation();
    if(cache_line) free(cache_line);
    return status;
  }
  
  debugmsg(4,"received ssh_user %s", user_str);
  status->ssh_user = malloc((strlen(user_str)+1)*sizeof(char));
  assert(status->ssh_user != NULL);
  strcpy(status->ssh_user, user_str);
  
  free(user_str);
  
  debugmsg(4,"received expire string %s", expire_str);
  /* status->expire = (time_t)120; */ /* TMP!!! */
  status->expire = (time_t)(atol(expire_str))+now;
  free(expire_str);
  free(status_str);
  
  /* Close the socket */
  globus_io_close(&conn_handle);
  
  /* Deactivate the Globus modules */
  globus_io_secure_authorization_data_destroy(&auth_data);
  globus_io_tcpattr_destroy(&attr);
  globus_deactivation();
  

  /* Auth done - update cache */

  if(conf->use_cache) {
    debugmsg(6,"Removing any old conflicting cache entries from cache file %s", cache_filename);
    /* We've still got the old line if there */
    if(cache_line != NULL) {
      debugmsg(8,"Removing conflicting cache entry %s from cache file %s", cache_line, cache_filename);
      remove_line(cache_line, cache_filename);
      free(cache_line);
    }
    
    debugmsg(6,"writing updated auth info to cache file %s", cache_filename);
    /* Cache format is host:port:userid:expire */
    debugmsg(4,"writing %s;%s;%s;%ld to cache file",status->server, status->port, status->ssh_user,(unsigned long int)status->expire);
    /* using 3*sizeof(tie_t) as an upper bound on the number of chars in expire */
    cache_line = malloc((strlen(status->server)+1+strlen(status->port)+1+strlen(status->port)+1+strlen(status->ssh_user)+1+3*sizeof(time_t)+1)*sizeof(char));
    assert(cache_line != NULL);
    sprintf(cache_line, "%s;%s;%s;%ld",status->server, status->port, status->ssh_user,(unsigned long int)(status->expire));
    insert_line(cache_line, cache_filename);
    free(cache_filename);
  }

  return status;
}
