/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 *
 * Functions for manipulating files on a line basis
 *
 */

#include "file_io.h"
#include "log.h"
#include "runtime_conf.h"
#include <globus_common.h> /* mutex */
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>

globus_mutex_t file_mutex;

int initialize_file_io()
{
  globus_mutex_init(&file_mutex, NULL);
  return 0;
}

int finalize_file_io()
{
  globus_mutex_destroy(&file_mutex);
  return 0;
}

char *match_line(char *search, char *filename)
{
  char *match = NULL;
  FILE *file;
  char *cur_line = NULL;
  size_t cur_linelen = 0;

  int found_match = 0;
  char *pos = NULL;
  
  globus_mutex_lock(&file_mutex);

  /* Open file in read and append mode */
  if((file = fopen(filename, "a+")) != NULL) {
    debugmsg(6,"Reading lines from file: %s (%x)", filename, file);
    while(getline(&cur_line, &cur_linelen, file) != -1) {
      while((pos = strrchr(cur_line,'\n')) != NULL) {
	/* truncate possible getline junk after newline */
	debugmsg(7,"removing newline junk from imported line pos %x", pos);
	memset(pos, 0, (cur_line+cur_linelen)-pos);
	*pos = '\0';
      }
      debugmsg(9,"tail is: %s",&(cur_line[cur_linelen-50]));
      debugmsg(9,"%s",cur_line);
      if(!strncmp(cur_line, search,strlen(search))) {
	found_match = 1;
	debugmsg(5,"found matching line %s in file",cur_line);
	break;
      }
    }
    
    if(found_match) {
      match = malloc((strlen(cur_line)+1)*sizeof(char));
      strcpy(match,cur_line);
    }
    
    if(cur_line) free(cur_line);

    fclose(file);
  } else {
    errmsg("opening file (%s) failed - check permissions!", filename);
  }

  globus_mutex_unlock(&file_mutex);

  return match;  
}

int insert_line(char *line, char *filename)
{
  FILE *file;
  char *cur_line = NULL;
  size_t cur_linelen = 0;

  int inserted_line = 0;
  char *pos = NULL;

  globus_mutex_lock(&file_mutex);

  /* Open file in read and append mode */
  if((file = fopen(filename, "a+")) != NULL) {
  
    debugmsg(6,"Reading lines from file: %s (%x)", filename, file);
    while(getline(&cur_line, &cur_linelen, file) != -1) {
      while((pos = strrchr(cur_line,'\n')) != NULL) {
	/* truncate possible getline junk after newline */
	debugmsg(7,"removing newline junk from imported line pos %x", pos);
	memset(pos, 0, (cur_line+cur_linelen)-pos);
	*pos = '\0';
      }
      debugmsg(9,"tail is: %s",&(cur_line[cur_linelen-50]));
      debugmsg(9,"%s",cur_line);
      if(!strcmp(cur_line, line)) {
	/* We could discriminate here and only insert if not there already.
	 However always inserting eases expiry of multiple concurrent sessions. */
	debugmsg(5,"found line in file");
	break;
      }
    }
    
    if (!inserted_line) {
      debugmsg(5,"inserting line in file %s", filename);
      if(fprintf(file, "%s\n", line) > 0) {
	inserted_line = 1;
        debugmsg(3,"inserted line %s in file %s", line, filename);
      } else {
	errmsg("writing line to file (%s) failed", filename);
	inserted_line = -2;
      }	
    }
    
    if(cur_line) free(cur_line);

    fclose(file);
  } else {
    errmsg("opening file (%s) failed - check permissions!", filename);
    inserted_line = -1;
  }

  globus_mutex_unlock(&file_mutex);

  return inserted_line;
}

int remove_line(char *line, char *filename)
{
  /* Remove first occurence of line from file */
  FILE *file = NULL, *tmp_file = NULL;
  int tmp_fd;
  char *tmp_filename = NULL;
  char *cur_line = NULL;
  size_t cur_linelen = 0;

  int removed_line = 0;
  char *pos = NULL;

  mode_t old_umask;
  struct stat buf;

  debugmsg(4,"removing line %s from file %s", line, filename);

  debugmsg(7,"setting umask to %d", 022);
  /* restrict tmp_file access to mode 644 (rw-r--r--) */
  old_umask = umask(022);

  /* If file doesn't exist we shouldn't get here - ok to return error */
  if(stat(filename, &buf) != 0) {
    errmsg("failed to stat file (%s) - line not removed", filename);
    return removed_line;
  }

  /* Create temporary copy without line in same dir as file */
  /* temp file in homedir should pretty much prevent attacks */
  tmp_filename = malloc((strlen(filename)+1+6+1)*sizeof(char));
  assert(tmp_filename != NULL);

  sprintf(tmp_filename, "%s.XXXXXX", filename);

  debugmsg(7,"build uninitialized tmp_filename %s", tmp_filename);
  
  if(mkstemp(tmp_filename) == -1) {
    errmsg("creating unique tmp file (%s) failed", tmp_filename);
  }

  debugmsg(7,"initialized tmp_filename to %s", tmp_filename);

  globus_mutex_lock(&file_mutex);
  
  /* Open file in read-only mode */
  if(! (file = fopen(filename, "r"))) {
    errmsg("open file (%s) failed", filename);
    free(tmp_filename);
    fclose(file);
    globus_mutex_unlock(&file_mutex);
    return removed_line;
  }

  /* create or truncate temp file if it exists */
  
  debugmsg(7,"opening tmpfile");
  if(((tmp_fd = open(tmp_filename, O_CREAT | O_TRUNC | O_RDWR)) == -1) ||\
     (tmp_file = fdopen(tmp_fd, "w")) == NULL) {
    errmsg("open tmp file (%s) failed", tmp_filename);
    free(tmp_filename);
    fclose(file);
    fclose(tmp_file);
    close(tmp_fd);
    globus_mutex_unlock(&file_mutex);
    return removed_line;
  }
  
  debugmsg(7,"reading lines from file: %s (%x)", filename, file);
  while(getline(&cur_line, &cur_linelen, file) != -1) {
    while((pos = strrchr(cur_line,'\n')) != NULL) {
      /* truncate possible getline junk after newline */
      debugmsg(7,"removing newline junk from imported line pos %x", pos);
      memset(pos, 0, (cur_line+cur_linelen)-pos);
      *pos = '\0';
    }
    debugmsg(9,"tail is: %s",&(cur_line[cur_linelen-50]));
    debugmsg(9,"%s",cur_line);
    if(!removed_line && !strcmp(cur_line, line)) {
      debugmsg(4,"not copying line-to-be-removed to new file");
      removed_line = 1;
    } else {
      debugmsg(4,"copying other line to new file %s", tmp_filename);
      fprintf(tmp_file,"%s\n",cur_line);
    }
  }
  
  fclose(file);

  debugmsg(7,"setting permissions %d on tmp_file", old_umask);
  /* set permissions before closing to prevent timing attacks */
  umask(old_umask);
  fchmod(tmp_fd, buf.st_mode);

  debugmsg(5,"uid %d - fchown file to %d:%d if root", conf->uid, buf.st_uid, buf.st_gid);
  if(conf->uid == 0) {
    if(fchown(tmp_fd, buf.st_uid, buf.st_gid) != 0) {
      errmsg("failed to change owner and group of tmp file (%s)", tmp_filename);
    }
  }

  debugmsg(7,"closing tmp_file");
  fclose(tmp_file);
  close(tmp_fd);

  debugmsg(7,"truncating old file %s with %s", filename, tmp_filename);
  if(rename(tmp_filename, filename) != 0) {
    errmsg("Truncating file failed");
    removed_line = 0;
  }

  globus_mutex_unlock(&file_mutex);

  free(cur_line);
  free(tmp_filename);

  return removed_line;
}
