/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
*/


/*
  
Gridssh gatekeeper server using globus I/O with GSI
authentication, gridmap authorization and encryption included.

Built upon the echoserver example by Arto Ters' (arto.teras@hip.fi)
from para2002 (http://www.csc.fi/para2002/tutorial.phtml).

The server accepts connections that authenticate with a valid grid-proxy.
Authorization is done if a matching unix user can be found in the gridmap file.
On acceptance a public key matching the private key (userkey.pem) of the user 
will be added to the ssh authorized_keys of that grid-mapped unix user.
In effect allowing the user to ssh-login with the private key.

In priciple it would be possible to use one of the negotiated passwordless 
keypairs instead, but OpenSSH requires more bits than they contain and stolen 
proxies/keys may still be too much of a security issue anyway.

By default the server also handles autoexpiration of the authorization in
accordance with the expiration of the proxy.

*/


#include "runtime_conf.h"
#include "gridsshd.h"
#include "log.h"
#include "key_handling.h"
#include "file_io.h"
#include "home_dir.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* basename */
#include <libgen.h>

/* signal handling */
#include <signal.h>
#include <sys/wait.h>

/* File manipulation */
#include <sys/types.h>
#include <sys/stat.h>

/* getopt */
#include <unistd.h>


/* getopt prototypes may not be included in unistd */
int getopt(int argc, char * const argv[], const char *optstring);
extern char *optarg;
extern int optind, opterr, optopt;

typedef struct thread_arguments {
  void *list_handle;
  void *conn_handle;
  void *attr;
} thread_arg;

/*  Global constants  */
extern runtime_conf *conf;
extern expire_queue *expire_q;

globus_mutex_t run_mutex;
globus_cond_t stop_running;
globus_mutex_t listen_mutex;
globus_cond_t got_connection;
globus_mutex_t threads_mutex;
/* Variables for security functionality - global for finalize access */
globus_io_attr_t attr;
globus_io_secure_authorization_data_t auth_data;
globus_io_handle_t list_handle;

volatile sig_atomic_t run = 1;
volatile int threads = 0;

int initialize_conf(int argc, char **argv)
{
  conf = malloc(sizeof(runtime_conf));
  assert(conf != NULL);
  
  /* Initialize runtime configuration structure, conf */
  conf->argv = argv;
  conf->argc = argc;
  conf->max_procs = DEFAULT_MAX_PROCS;
  conf->log_stats = 0;
  conf->use_cache = 1;
  conf->debug_level = 0;
  conf->sshd_port = (char *)UNUSED;
  conf->gridsshd_port = DEFAULT_GRIDSSHD_PORT;
  conf->address = 0;
  conf->uid = getuid();
  conf->gid = getgid();
  conf->pid = getpid();
  conf->pid_filename = NULL;
  conf->prog = basename(argv[0]);
  assert(conf->prog != NULL);
  conf->auth_filename = DEFAULT_AUTH_FILENAME;
  conf->expire_mode = DEFAULT_EXPIRE_MODE;
  conf->out_file = NULL;
  conf->err_file = NULL;
  conf->daemon_mode = 1;
  conf->forward_proxy = 0;

  return 0;
}

int finalize_conf(void)
{
  conf->argv = NULL;
  conf->prog = NULL;
  conf->pid_filename = NULL;
  conf->auth_filename = NULL;
  free(conf);
  return 0;
}

int inc_active_threads(globus_mutex_t *threads_mutex)
{
  int t = 0;

  if(globus_mutex_lock(threads_mutex) != 0) {
    errmsg("%s: locking threads_mutex failed", __FUNCTION__);
  }

  t = ++threads;

  if(globus_mutex_unlock(threads_mutex) != 0) {
    errmsg("%s: unlocking threads_mutex failed", __FUNCTION__);
  }
  
  return t;
}

int dec_active_threads(globus_mutex_t *threads_mutex)
{
  int t = 0;

  if(globus_mutex_lock(threads_mutex) != 0) {
    errmsg("%s: locking threads_mutex failed", __FUNCTION__);
  }

  t = --threads;

  if(globus_mutex_unlock(threads_mutex) != 0) {
    errmsg("%s: unlocking threads_mutex failed", __FUNCTION__);
  }
  
  return t;
}

int get_active_threads(globus_mutex_t *threads_mutex)
{
  int t = 0;

  if(globus_mutex_lock(threads_mutex) != 0) {
    errmsg("%s: locking threads_mutex failed", __FUNCTION__);
  }

  t = threads;

  if(globus_mutex_unlock(threads_mutex) != 0) {
    errmsg("%s: unlocking threads_mutex failed", __FUNCTION__);
  }
  
  return t;
}

int initialize_threads(globus_mutex_t *threads_mutex)
{
  debugmsg(5,"%s setting up mutex", __FUNCTION__);

  if(globus_mutex_init(threads_mutex, NULL) != 0) {
    errmsg("%s: initializing threads_mutex failed", __FUNCTION__);
  }

  threads = 0;

  return 0;
}

int finalize_threads(globus_mutex_t *threads_mutex)
{
  int active = -1;

  debugmsg(4,"%s waiting for threads to finish", __FUNCTION__);

  /* TODO: add max-tries? */
  while((active = get_active_threads(threads_mutex)) > 0) {
    debugmsg(7,"%s waiting for %d threads to finish", __FUNCTION__);
  }
  debugmsg(8,"%s last thread finished", __FUNCTION__);
  
  if(globus_mutex_destroy(threads_mutex) != 0) {
    errmsg("%s: destroying threads_mutex failed", __FUNCTION__);
  }
  
  debugmsg(5,"%s done waiting", __FUNCTION__);
  
  return 0;
}

void sig_handler(int sig)
{
  int pid = getpid();

  debugmsg(7,"%d caught sig %d", pid, sig);
  
  if(sig != SIGTERM && sig != SIGINT) {
    debugmsg(2,"signal %d ignored", sig);
    return;
  }

  if(pid == conf->pid) {
    if(run == 1) {
      debugmsg(4,"main process got sig %d", sig);
      run = 0;
      if(globus_cond_broadcast(&stop_running) != 0) {
	errmsg("%s: broadcasting stop_running condition failed", __FUNCTION__);
      }
    }
  } else if (getppid() == conf->pid) {
    debugmsg(5,"main child caught sig %d", sig);
    run = 0;
    if(globus_cond_broadcast(&got_connection) != 0) {
      errmsg("%s: broadcasting got_connection condition failed", __FUNCTION__);
    }
  } else {
    debugmsg(5,"grandchild caught sig %d", sig);
    run = 0;
    exit(0);
  }
}


int finalize(int level)
{

  if(run == 1 && level > 2) {
    debugmsg(4,"child process %d called finalize directly - notify main", getpid());
    if(kill(conf->pid, SIGTERM) != 0) {
      errmsg("Sending TERM signal to %d failed", conf->pid);
    }
    /* don't wait for signal to arrive:
       -if threads are running we'll wait for them below (implicit wait for signal)
       -else we can exit as soon as we've cleaned up 
    */
  }

  debugmsg(4,"process %d finalizing", getpid());

  /* begin at specified level and fall through all lower levels */
  switch(level) {
  case 11:
    debugmsg(4,"waiting for threads to return");
    finalize_threads(&threads_mutex);    
  case 10:
    debugmsg(4,"destroying auth data");
    globus_io_secure_authorization_data_destroy(&auth_data);
  case 9:
    debugmsg(4,"destroying attr data");
    globus_io_tcpattr_destroy(&attr);
  case 8:
    debugmsg(4,"closing list_handle");
    globus_io_cancel(&list_handle, GLOBUS_FALSE);
    globus_io_close(&list_handle);
  case 7:
    debugmsg(4,"destroying run cond synch structures");
    globus_mutex_destroy(&run_mutex);
    globus_cond_destroy(&stop_running);
  case 6:
    debugmsg(4,"destroying listen cond synch structures");
    globus_mutex_destroy(&listen_mutex);
    globus_cond_destroy(&got_connection);
  case 5:
    debugmsg(4,"emptying expire queue");
    finalize_expire_q();
  case 4:
    debugmsg(4,"finalizing file io");
    finalize_file_io();
  case 3:
    debugmsg(4,"deactivating globus modules");
    globus_deactivation();
  case 2:
    debugmsg(4,"%d finalizing log",getpid());
    finalize_log();
  case 1:
    finalize_conf();
  default:
    errmsg("unknown level (%d) passed to finalize", level);
  }

  return 0;
}

int globus_activation(void)
{
  int err = 0;

  debugmsg(4,"%d activating globus modules",getpid());

  /* Activating all modules that we rely on directly. This may not be strictly 
     necessary as globus does handle module dependencies to some degree. */

  if (globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus common module activation failed");
    err++;
  }
  
  if(globus_module_activate(GLOBUS_OPENSSL_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus openssl module activation failed");
    err++;
  }
  
  if(globus_module_activate(GLOBUS_GSI_GSS_ASSIST_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus assist module activation failed");
    err++;
  }

  if(globus_module_activate(GLOBUS_GSI_CREDENTIAL_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus credential module activation failed");
    err++;
  }

  if (globus_module_activate(GLOBUS_IO_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus I/O module activation failed");
    err++;
  }

  return err;
}


int globus_deactivation(void)
{
  int err = 0;

  debugmsg(4,"%d deactivating globus modules",getpid());

  debugmsg(8,"%d deactivating globus io module",getpid());
  if (globus_module_deactivate(GLOBUS_IO_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus I/O module deactivation failed");
    err++;
  }
  
  debugmsg(8,"%d deactivating globus credential module",getpid());
  if(globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus credential module deactivation failed");
    err++;
  }

  debugmsg(8,"%d deactivating globus assist module",getpid());
  if(globus_module_deactivate(GLOBUS_GSI_GSS_ASSIST_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus assist module deactivation failed");
    err++;
  }

  if(globus_module_deactivate(GLOBUS_OPENSSL_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus openssl module deactivation failed");
    err++;
  }

  if (globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS) {
    errmsg("globus common module activation failed");
    err++;
  }

  debugmsg(4,"%d done deactivating globus modules",getpid());

  return err;
}


void listen_callback(void *callback_arg, globus_io_handle_t *conn_handle, globus_result_t result)
{
  int *new_connection = callback_arg;

  debugmsg(8,"listen callback");

  if (result != GLOBUS_SUCCESS) {
    /* listen failed - probably lack of authentication */
    errmsg("listen failed - ignoring");
    *new_connection = 0;
  } else {
    debugmsg(5,"listen succeeded - allow accept");
    *new_connection = 1;
  }

  /* wake up listening child */
  if(globus_cond_broadcast(&got_connection) != 0) {
    errmsg("%s: broadcasting got_connection condition failed", __FUNCTION__);
  }

  return;
}

globus_bool_t auth_callback(void *arg, globus_io_handle_t *handle, globus_result_t res, char *identity, gss_ctx_id_t context)
{
  char *username;
  int ret;

  debugmsg(4,"child (%d) handling GSI authorization of %s",getpid(),identity);
  debugmsg(8,"handling client with handle %d ...",*handle);

  /* if GRIDMAP environment var is not already set - do so since it is 
     required by gridmap lookup */
  if(setenv("GRIDMAP", DEFAULT_GRIDMAP_FILE,0) != 0) {
    errmsg("%s: setting GRIDMAP environment failed", __FUNCTION__);
  }

  debugmsg(3,"using gridmap in %s", getenv("GRIDMAP"));
  
  ret = globus_gss_assist_gridmap(identity, &username);

  if (ret == 0) {
    debugmsg(4,"succesfully authorized %s against gridmap user %s", identity, username);
    ret = GLOBUS_TRUE;
  } else {
    errmsg("authorizing %s against gridmap failed (%d)", identity, ret);
    ret = GLOBUS_FALSE;
  }
  
  /* we can't extract delegated proxy and insert key here - wait for session */

  if (username) free(username);
  
  return ret;
}


void *expire_thread(void *thread_args)
{
  int expired;
  int i;

  inc_active_threads(&threads_mutex);

  while(run) {
    debugmsg(8,"expire_thread calling check_expired");
    expired = check_expire();
    debugmsg(4,"expire_thread expired %d keys", expired);
    /* responsive wait loop */
    for(i = 0; i < DEFAULT_EXPIRE_INTERVAL; i++) {
      if(run) sleep(2);
    }
  }
  
  debugmsg(5,"expire_thread leaving");

  dec_active_threads(&threads_mutex);

  return (void *)NULL;
}


void *thread_session(void *thread_args)
{
  globus_result_t rptr;

  thread_arg *args = (thread_arg *)thread_args;

  inc_active_threads(&threads_mutex);

  debugmsg(4,"thread handling client");
  handle_session(args->conn_handle);
  
  debugmsg(5,"closing conn_handle");
  rptr = globus_io_close(args->conn_handle);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("closing the connected socket failed");
  }

  free(args->conn_handle);
  args->conn_handle = NULL;

  free(args);
  
  debugmsg(4,"session thread leaving");

  dec_active_threads(&threads_mutex);

  return (void *)NULL;
}

void handle_session(globus_io_handle_t *conn_handle)
{
  /* Delegation variables*/
  globus_result_t rptr;
  gss_cred_id_t cred;
  char *deleg_proxy_filename = NULL;
  OM_uint32 major_status;
  OM_uint32 minor_status;
  gss_name_t server_name = GSS_C_NO_NAME;
  gss_buffer_desc server_buffer_desc = GSS_C_EMPTY_BUFFER;
  OM_uint32 lifetime = 0;
  gss_buffer_desc deleg_proxy_desc;
  char *cp;

  char *openssh_pubkey = NULL;

  /* auth_file */
  char *unix_user = NULL, *home_dir = NULL;

  char *abs_auth_filename = NULL;

  globus_size_t nbytes;
  char *status_str = NULL;

  int key_inserted = 0;
  time_t now = 0;

  debugmsg(5,"child (%d) handling client with handle %d ...",getpid(),*conn_handle);

  rptr = globus_io_tcp_get_delegated_credential(conn_handle, &cred);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("extracting the delegated proxy credential failed");
    return;
  }
  
  major_status = gss_inquire_cred(&minor_status, cred, &server_name,
				  &lifetime, NULL, NULL);
  if (major_status == GSS_S_COMPLETE) {
    major_status = gss_display_name(&minor_status, server_name,
				    (gss_buffer_t)(&server_buffer_desc), NULL);
    gss_release_name(&minor_status, &server_name);
  } else {
    errmsg("inquiry of delegated proxy credential failed");
    return;
  }
  
  debugmsg(2,"thread %d handling OpenSSH authorization request from \"%s\"", globus_thread_self(), server_buffer_desc.value);
  debugmsg(6,"delegated lifetime: %d",lifetime);
  
  switch(conf->expire_mode) {
  case NEVER_EXPIRE:
    lifetime = MAX_LIFETIME;
    break;
  case PROXY_EXPIRE:
    /* use credential value from above */
    break;
  case AGGRESSIVE_EXPIRE:
    lifetime = DEFAULT_LOGIN_TIMEOUT/1000000;
    break;
  }

  /* TODO: don't export at all? */
  /* 
     gss_cred_id_t cred = [something];
     gss_cred_id_desc *cred_desc = NULL;
     globus_gsi_cred_handle_t gsi_cred_handle;
     cred_desc = (gss_cred_id_desc *) cred;
     gsi_cred_handle = cred_desc->cred_handle;
  */
  major_status = gss_export_cred(&minor_status, cred, NULL, 1,
				 &deleg_proxy_desc);
  
  /* deleg_proxy_desc.value contains 
     "X509_USER_PROXY=[delegated proxy filename]" */
  if (major_status == GSS_S_COMPLETE) {
    debugmsg(4,"delegated proxy string: \"%s\"",(char *)deleg_proxy_desc.value);
    cp = strchr((char *)deleg_proxy_desc.value, '=');
    if(cp == NULL) {
      errmsg("did not find expected '=' in delegated proxy string %s", (char *)deleg_proxy_desc.value);
      return;
    }
    *cp = '\0';
    cp++;
    /* Now cp points at the filename of the delegated proxy 
       and deleg_proxy_desc.value only contains the env. variable name */
    if(strlen(cp) <= 0) {
      errmsg("no delegated proxy filename found in delegated proxy string %s", (char *)deleg_proxy_desc.value);
      return;
    }
    /* add one to include string termination */
    deleg_proxy_filename = (char *)malloc(strlen(cp)+1);
    assert(deleg_proxy_filename != NULL);
    strcpy(deleg_proxy_filename, cp);
  } else {
    errmsg("exporting delegated proxy credential failed");
    unlink(deleg_proxy_filename);
    free(deleg_proxy_filename);
    return;
  }
  
  /* Find home dir of grid mapped user */
  if(globus_gss_assist_gridmap((char *)(server_buffer_desc.value), &unix_user) != 0) {
    errmsg("couldn't extract gridmap user");
    free(unix_user);
    unlink(deleg_proxy_filename);
    free(deleg_proxy_filename);
    return;
  }

  debugmsg(6,"retrieving homedir for user %s", unix_user);
  if((home_dir = get_user_home(unix_user, (void *)-1)) == NULL) {
    errmsg("couldn't extract home dir for user %s - can't autorize user %s", unix_user, (char *)(server_buffer_desc.value));
    errmsg("please make sure user %s exists as a valid unix user or change the gridmap file to map %s to a valid local user", unix_user, (char *)(server_buffer_desc.value));
    
    free(unix_user);
    unlink(deleg_proxy_filename);
    free(deleg_proxy_filename);
    return;
  }

  debugmsg(4,"found homedir %s for user %s", home_dir, unix_user);

  /* Allocate space for homedir/auth_filename\0 string */
  abs_auth_filename = malloc((strlen(home_dir)+1+strlen(conf->auth_filename)+1)*sizeof(char));
  assert(abs_auth_filename != NULL);
  
  debugmsg(5,"building abs_auth_filename of len %d",(strlen(home_dir)+1+strlen(conf->auth_filename)+1)*sizeof(char));
  sprintf(abs_auth_filename, "%s/%s",home_dir,conf->auth_filename);

  debugmsg(5,"built abs_auth_filename %s",abs_auth_filename);

  free(home_dir);

  debugmsg(4,"handling key extraction %s",deleg_proxy_filename);
  openssh_pubkey = extract_openssh_pubkey(deleg_proxy_filename);
  if(conf->forward_proxy) {
	    debugmsg(3,"Delegated proxy file %s stored for forwarding", deleg_proxy_filename);
  } else {
 	 if(unlink(deleg_proxy_filename) != 0) {
	    errmsg("unlinking delegated proxy file %s failed", deleg_proxy_filename);
	 }
  }	 
  free(deleg_proxy_filename);
  if(openssh_pubkey == NULL) {
    errmsg("openssh public key extraction failed");
    free(abs_auth_filename);
    return;
  }

  key_inserted = insert_line(openssh_pubkey, abs_auth_filename);

  debugmsg(6,"building status response for client");
    /* the number of digits in time_t is always <= 3 chars per byte */
  status_str = malloc((strlen(SUCCESS_STR)+1+strlen(unix_user)+1+3*sizeof(time_t)+1)*sizeof(char));
  assert(status_str != NULL);

  /* Make sure MAX_STATUS_LEN is big enough to hold the string */
  /* ..client must be able to receive the string witout knowing the length */
  assert(strlen(status_str) < MAX_STATUS_LEN);

  if(key_inserted < 0) {
    sprintf(status_str,"%s;;",FAILED_STR);
  } else {
    debugmsg(7,"status: %s, %s, %ld",SUCCESS_STR, unix_user, (unsigned long int)lifetime);
    sprintf(status_str,"%s;%s;%ld",SUCCESS_STR, unix_user, (unsigned long int)lifetime);
    debugmsg(8,"status string written");
  }

  debugmsg(8,"freeing unix_user");
  free(unix_user);
  debugmsg(8,"freed unix_user");

  debugmsg(4,"sending status response %s to client", status_str);
  /* TODO: check result of write! */
  globus_io_write(conn_handle, (globus_byte_t *)status_str, strlen(status_str), &nbytes);
  debugmsg(7,"sent status response to client");

  now = time(NULL);
  assert(now != ((time_t) -1));

  if(key_inserted == 1) {
    switch(conf->expire_mode) {
    case NEVER_EXPIRE:
      debugmsg(4,"never expire mode (%d) - nothing to do here",conf->expire_mode);
      break;
    case PROXY_EXPIRE:
      debugmsg(5,"proxy expire mode (%d) - scheduling key for later removal",conf->expire_mode);
      debugmsg(3,"time is now %d scheduling removal in %d seconds",now, lifetime);
      sched_expire(openssh_pubkey, abs_auth_filename,now+lifetime);
      break;
    case AGGRESSIVE_EXPIRE:
      debugmsg(4,"aggresive expire mode (%d) - scheduling removal of key in %d seconds",conf->expire_mode, lifetime);
      debugmsg(3,"time is now %d scheduling removal in %d seconds",now, lifetime);
      sched_expire(openssh_pubkey, abs_auth_filename,now+lifetime);
      break;
    default:
      warnmsg(0,"unknown expire mode (%d) - not yet supported - falling back to default",conf->expire_mode);
    }
  }
  
  free(openssh_pubkey);
  free(abs_auth_filename);

  if(key_inserted < 0) {
    debugmsg(2,"thread %d finished session with \"%s\" without authorizing", globus_thread_self(), (char *)(server_buffer_desc.value));
  } else {
    debugmsg(2,"thread %d authorized and finished session with \"%s\"", globus_thread_self(), (char *)(server_buffer_desc.value));
  }
}


char *extract_openssh_pubkey(char *deleg_proxy_filename)
{
  globus_result_t rptr;
  char *openssh_pubkey = NULL;
  globus_gsi_cred_handle_t proxy_cred = NULL;
  X509 *user_cert = NULL;
  STACK_OF(X509) *px509_chain = NULL;
  int chain_len = -1;
  EVP_PKEY *x509_pubkey = NULL;
  Key *pubkey = NULL;

  rptr = globus_gsi_cred_handle_init (&proxy_cred, NULL);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("creating credential handle.");
    globus_gsi_cred_handle_destroy(proxy_cred);
    return NULL;
  }
	  
  /* TODO: security: if we have to read we should at least check owner,perms */
  rptr = globus_gsi_cred_read_proxy(proxy_cred, deleg_proxy_filename);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("reading proxy.");
    globus_gsi_cred_handle_destroy(proxy_cred);
    return NULL;
  }

	  
  rptr = globus_gsi_cred_get_cert_chain(proxy_cred, &px509_chain);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("getting certificate chain.");
    free(px509_chain);
    globus_gsi_cred_handle_destroy(proxy_cred);
    return NULL;
  }
	  
  chain_len = sk_X509_num(px509_chain);

  if(chain_len < 2) {
    errmsg("certificate chain too short - can't extract pubkey.");
    free(px509_chain);
    globus_gsi_cred_handle_destroy(proxy_cred);
    return NULL;
  }

  debugmsg(4,"extracted certificate chain has %d elements",chain_len);

  /* Last certificate but one in chain is the original user_cert */
  user_cert = sk_X509_value(px509_chain, chain_len-2);

  if ((x509_pubkey = X509_get_pubkey(user_cert)) == NULL) {
    errmsg("getting pub key.");
    user_cert = NULL;
    free(px509_chain);
    globus_gsi_cred_handle_destroy(proxy_cred);
    return NULL;
  }
  
  debugmsg(9,"initializing openssh key from pkey %x", x509_pubkey->pkey);
  pubkey = key_init(KEY_RSA, (void *)((x509_pubkey->pkey).rsa));
  if(pubkey == NULL) {
    errmsg("creating pub key string representation failed.");
    user_cert = NULL;
    free(px509_chain);
    globus_gsi_cred_handle_destroy(proxy_cred);
    return NULL;
  }
  
  debugmsg(8,"writing pubkey to string");
  key_write_str(pubkey, &openssh_pubkey);
  debugmsg(5,"wrote pubkey to string: %s", openssh_pubkey);

  key_free(pubkey);

  rptr = globus_gsi_cred_handle_destroy(proxy_cred);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("destroying proxy credential handle.");
  }

  user_cert = NULL;
  free(x509_pubkey);
  free(px509_chain);

  return openssh_pubkey;
}


void print_usage(void)
{
  showstr("### %s %s ###\n", conf->prog, GRIDSSHD_VERSION);
  showstr("Usage: %s [options]\n", conf->prog);
  showstr("Options:\n");
  showstr("\t-a auth_file: specify that sshd authorizes user against ~user/auth_file\n\t\trather than ~user/%s\n", DEFAULT_AUTH_FILENAME);
  showstr("\t-d debug_level: select debug level (default 0 i.e. debug off)\n");
  showstr("\t-e expire_mode: select expire mode (default %d)\n", DEFAULT_EXPIRE_MODE);
  showstr("\t\t0: do not expire at all - authorized users get permanent access\n");
  showstr("\t\t1: expire when proxy expires or when server halts\n");
  showstr("\t\t2: expire after %d seconds when login would have timed out\n", DEFAULT_LOGIN_TIMEOUT/1000000);
  showstr("\t-h: Print this help text\n");
  showstr("\t-o outfile: send output to outfile instead of stdout\n");
  showstr("\t-p port: force gridsshd to bind to the specified port (default %d)\n", DEFAULT_GRIDSSHD_PORT);
  showstr("\t-v: Print version details and exit\n");
  showstr("\nExample: %s -d 3 -p 1234\n", conf->prog);
}


int main(int argc, char *argv[]) {
  int ch; /* used to hold command line option */
  int itmp;
  const char *optstr = "a:d:De:fho:p:v";
  char *log_filename = NULL;
  char *home = NULL;
  char *pid_str = NULL;
  pid_t id;

  int error, i, status, new_connection;

  unsigned short int port = -1;
  int retries_left;

  /* Variables for globus functions */
  int rval;
  globus_result_t rptr;
  globus_thread_t tid;
  globus_thread_t expire_tid;

  thread_arg *thread_args = NULL;

  initialize_conf(argc, argv);
  /* output to stdout and stderr so far */
  initialize_log(NULL, NULL);

  /****************************************************************************/
  /*** Don't do anything before this point as we need conf for most actions ***/
  /****************************************************************************/

  /* Establish a handler for SIGTERM and SIGINT signals. */
  if(signal(SIGTERM, sig_handler) == SIG_ERR) {
    showstr("creating TERM signal handler failed\n");
    finalize(2);
    exit(EXIT_FAILURE);
  }
  if(signal(SIGINT, sig_handler) == SIG_ERR) {
    showstr("creating INT signal handler failed\n");
    finalize(2);
    exit(EXIT_FAILURE);
  }

  /* get args */
  /* scheduler uses original argv, argc */
  while (-1 != (int) (ch = getopt(argc, argv, optstr))) {
    switch (ch) {
    case 'a':
      conf->auth_filename = optarg;
      break;
    case 'h':
      print_usage();
      finalize(2);
      exit(EXIT_SUCCESS);
    case 'd':
      itmp = atoi(optarg);
      if (itmp >= 0)
	conf->debug_level = itmp;
      else {
	showstr("Illegal argument to -d option: debug_lvl must be a nonnegative integer!\n");
	print_usage();
	finalize(2);
	exit(EXIT_FAILURE);
      }
      break;
    case 'D':
      conf->daemon_mode = 0;
      break;
    case 'e':
      itmp = atoi(optarg);
      if (itmp >= 0 && itmp < 3)
	conf->expire_mode = itmp;
      else {
	showstr("Illegal argument to -e option: expire_mode must be 0, 1 or 2!\n");
	showstr("\t0: do not expire at all - authorized users get permanent access\n");
	showstr("\t1: expire when proxy expires or when server halts\n");
	showstr("\t2: expire after %d seconds when login would have timed out (default)\n", DEFAULT_LOGIN_TIMEOUT/1000000);
	print_usage();
	finalize(2);
	exit(EXIT_FAILURE);
      }
      break;
    case 'f':
      conf->forward_proxy = 1;
      showstr("Enabling experimental proxy forward code. Please note that forwarding isn't fully implemented yet!\n");
      break;
    case 'o':
      log_filename = optarg;
      break;
    case 'p':
      itmp = atoi(optarg);
      if (itmp > 0 && itmp < MAX_PORT) {
	conf->gridsshd_port = optarg;
      } else {
	showstr("Illegal argument to -p option: gridsshd port must be a positive integer (< %d)!\n", MAX_PORT);
	print_usage();
	finalize(2);
	exit(EXIT_FAILURE);
      }
      break;
    case 'v':
      showstr("### %s %s ###\n", conf->prog, GRIDSSHD_VERSION);
      finalize(2);
      exit(EXIT_SUCCESS);
      break;
    case '?':
      /* getopt returns '?' if it finds a flag not listed in the options string */
      showstr("Unrecognized option: %c\n", optopt);
      print_usage();
      finalize(2);
      exit(EXIT_FAILURE);
    default:
      showstr("getopt error? condition unaccounted for?\n");
      finalize(2);
      exit(EXIT_FAILURE);
    }
  }
  
  if(conf->daemon_mode) {
    /* fork a detached child to continue as a daemon */
    id = fork();
    if (id < 0) exit(EXIT_FAILURE); /* fork error */
    if (id > 0) exit(EXIT_SUCCESS); /* parent exits */

    /* Obtain a new process group */
    setsid();

    /* update pid */
    conf->pid = getpid();
    debugmsg(5,"forked process set pid to %d.", conf->pid);

    /* Close all open file descriptors */
    for (i=getdtablesize();i>=0;--i) close(i);

    /* Redirect stdin(0), stdout(1) and stderr(2) to /dev/null */
    /* Unix assigns descriptors sequentially so this will work */
    i = open("/dev/null",O_RDWR); /* stdin */
    dup(i); /* stdout */
    dup(i); /* stderr */

    if(log_filename == NULL) {
      log_filename = DEFAULT_OUT_FILE;
    }
  }

  if(log_filename) {
    if(!initialize_log(log_filename, log_filename)) {
      finalize(2);
      exit(EXIT_FAILURE);
    }
  }

  showmsg("%s server version %s starting", conf->prog, GRIDSSHD_VERSION);
  
  if (conf->uid != 0) {
    debugmsg(3,"running server as a non-privileged user.");
    
    /* save pid in ~/gridsshd.pid */
    /* TODO: handle error instead of assert */
    home = get_home();
    assert(home != NULL);
    conf->pid_filename = malloc((strlen(home)+1+strlen(conf->prog)+4+1)*sizeof(char));
    assert(conf->pid_filename != NULL);
    sprintf(conf->pid_filename, "%s/%s.pid", home, conf->prog);
    free(home);

    if(!getenv("X509_USER_KEY") || !getenv("X509_USER_CERT")){
      warnmsg(0,"trying to run the server as a non-privileged user with at least one of the X509_USER_CERT and X509_USER_KEY environment variables unset.\nIf this causes authentication/authorization problems, please set them to point at a user readable host certificate and key pair or run the server as a privileged user like 'root'.");
    }
  } else {
    conf->pid_filename = DEFAULT_PID_FILE;
  }

  id = fork();
  if (id < 0) { 
    errmsg("fork failed (%d)", id);
    exit(EXIT_FAILURE);
  } else if (id > 0) {
    /* write pid file */
    /* pid never takes up more than 3 chars per byte */
    pid_str = malloc((3*sizeof(pid_t)+1)*sizeof(char));
    assert(pid_str != NULL);
    sprintf(pid_str, "%d", conf->pid);

    /* TODO: truncate file if already there */
    if(insert_line(pid_str, conf->pid_filename) != 1) {
      errmsg("Writing pid %s to %s failed - check permissions", pid_str, conf->pid_filename);
    }
    free(pid_str);
    if(conf->uid != 0) free(conf->pid_filename);

    /* parent idle waits for signals */
    while(run) {
      /* TODO: check return values */
      globus_mutex_lock(&run_mutex);
      debugmsg(4,"main process waiting for termination condition");
      globus_cond_wait(&stop_running, &run_mutex);
      debugmsg(2,"main process was told to terminate");
      globus_mutex_unlock(&run_mutex);
    }

    /* must have caught a signal to exit */
    debugmsg(5,"main process sending TERM to child");
    kill(id, SIGTERM);
    debugmsg(4,"main process caught signal - waiting for child %d", id);
    waitpid(id, &status, 0);
    debugmsg(3,"child %d returned with status %d", id, status);
    /* TODO: count and show number of handled request? */
    showmsg("%s server shutting down", conf->prog);
    finalize(2);
    return status;
  }

  /* child handles connections */

  /* Stop handling SIGINT in children and leave that to main process only */
  signal(SIGINT, SIG_IGN);
  
  /* TODO: initialize function like finalize? */

  rval = globus_activation();
  if (rval != GLOBUS_SUCCESS) {
    errmsg("globus module activation failed");
    finalize(3);
    exit(EXIT_FAILURE);
  }
  
  initialize_file_io();
  initialize_threads(&threads_mutex);
  initialize_expire_q();

  globus_mutex_init(&run_mutex, NULL);
  globus_cond_init(&stop_running, NULL);
  globus_mutex_init(&listen_mutex, NULL);
  globus_cond_init(&got_connection, NULL);
  
  /* Initialize the attributes structure. */
  rptr = globus_io_tcpattr_init(&attr);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("creating socket attribute failed");
      finalize(8);
      exit(EXIT_FAILURE);
  }
  
  /* Use GSS-API authentication */
  rptr = globus_io_attr_set_secure_authentication_mode(&attr, 
						       GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI, GSS_C_NO_CREDENTIAL);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("setting secure authentication mode failed");
    finalize(8);
    exit(EXIT_FAILURE);
  }
  
  /* Initialize the authorization data */
  rptr = globus_io_secure_authorization_data_initialize(&auth_data);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("initializing authorization data failed");
    finalize(9);
    exit(EXIT_FAILURE);
  }
  
  /* Set the authorization mode to use a gridmap lookup */
  rptr = globus_io_secure_authorization_data_set_callback(&auth_data, auth_callback, NULL);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("setting authorization callback failed");
    finalize(9);
    exit(EXIT_FAILURE);
  }
  
  rptr = globus_io_attr_set_secure_authorization_mode(&attr, GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK, &auth_data);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("setting authorization mode failed");
    finalize(9);
    exit(EXIT_FAILURE);
  }
  
  /* We transfer the public key using full proxy delegation */
  rptr = globus_io_attr_set_secure_delegation_mode(&attr, GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("setting delegation mode failed");
    finalize(9);
    exit(EXIT_FAILURE);
  }
  
  /* Set the channel mode to use wrapping */
  rptr = globus_io_attr_set_secure_channel_mode(&attr, 
						GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("setting channel mode failed");
    finalize(9);
    exit(EXIT_FAILURE);
  }
  
  /* Use encryption, SAFE instead of PRIVATE is also possible
     but gives only integrity protection */
  rptr = globus_io_attr_set_secure_protection_mode(&attr, 
						   GLOBUS_IO_SECURE_PROTECTION_MODE_PRIVATE);
  if (rptr != GLOBUS_SUCCESS) {
    errmsg("setting encrypted mode failed");
    finalize(9);
    exit(EXIT_FAILURE);
  }

  /*  Create the listening socket  */
  for (retries_left = DEFAULT_BIND_RETRIES; retries_left > 0; retries_left--) {
    debugmsg(8,"trying to create listening socket (%d retries left)", retries_left);
    port = (unsigned short) atoi(conf->gridsshd_port);
    rptr = globus_io_tcp_create_listener(&port, DEFAULT_MAX_CONNECTIONS, &attr, &list_handle);
    if (rptr == GLOBUS_SUCCESS) {
      debugmsg(8,"successfully created listening socket (%d retries left)", retries_left);
      break;
    } else {
      warnmsg(retries_left,"creating listening socket failed (%d retries left)",retries_left);
    }
    sleep(1);
  }
  
  if(retries_left > 0) {
      debugmsg(5,"created listening socket");
      debugmsg(9,"... with %d retries left", retries_left);
  } else {
      errmsg("creating listening socket on port %s failed - now giving up after %d retries",conf->gridsshd_port, DEFAULT_BIND_RETRIES);
      finalize(10);
      exit(EXIT_FAILURE);
  }


  if(conf->expire_mode != NEVER_EXPIRE) {
    debugmsg(4,"creating expire thread");
    globus_thread_create(&expire_tid,NULL,&expire_thread,(void *)NULL);
  } else {
    debugmsg(4,"never expire mode - not spawning expire thread");
  }
  
  /*  Enter an infinite loop to respond to client requests */
  error = 0;
  while (error == 0 && run) {
       debugmsg(5,"listening for connections");
       /* We can not block on listen as it will block clean shutdown */
       new_connection = 0;
       rptr = globus_io_tcp_register_listen(&list_handle, listen_callback, &new_connection);
       if (rptr != GLOBUS_SUCCESS) {
	 errmsg("calling listen");
	 error++;
	 continue;
       }

       globus_mutex_lock(&listen_mutex);
       debugmsg(4,"main child waiting for connection condition");
       globus_cond_wait(&got_connection, &listen_mutex);
       debugmsg(5,"main child woke up from connection wait");
       globus_mutex_unlock(&listen_mutex);

       if(!new_connection) { 
	 debugmsg(4,"main child got woken up but no connection");
	 continue;
       }

       debugmsg(4,"connection arrived");

       /* set up new thread to handle client */
       thread_args = malloc(sizeof(thread_arg));
       assert(thread_args != NULL);
       thread_args->list_handle = &list_handle;
       thread_args->conn_handle = malloc(sizeof(globus_io_handle_t));
       thread_args->attr = &attr;

       debugmsg(4,"accepting connection");

       rptr = globus_io_tcp_accept(thread_args->list_handle, thread_args->attr, thread_args->conn_handle);
       if (rptr != GLOBUS_SUCCESS) {
	 char *tmp;
	 tmp = globus_object_printable_to_string(globus_error_get(rptr));
	 warnmsg(0,"accept connection failed with client side error (%s) - ignoring", tmp);
	 
	 free(tmp);

	 free(thread_args->conn_handle);
	 free(thread_args);
	 continue;
       }
	
       /* globus doesn't support thread_join so we might as well truncate tid */
       debugmsg(5,"accepted connection - spawning thread to handle key session");
       globus_thread_create(&tid, NULL, &thread_session, (void *) thread_args);
  }

  if (error > 0) {
    errmsg("server terminating due to an error in the listening loop");
  }

  debugmsg(2,"main child finished loop - finalizing");

  finalize(11);

  /* child leaving */
  if (error) {
    exit(EXIT_FAILURE);
  }
  exit(EXIT_SUCCESS);
}
