#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <string>
#include <vector>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"
#include "Giis.h"

#include "ngui.h"


extern "C"
int ngstatxx (const vector <string> & jobs,
	      const vector <string> & joblists,
	      const vector <string> & clusterselect_,
	      const vector <string> & clusterreject_,
	      const vector <string> & status,
	      const bool all,
	      const vector <string> & giisurls,
	      const bool clusters,
	      const bool longlist,
	      int timeout,
	      int debug,
	      const bool anonymous) {

  vector <string> clusterselect = ResolveAliases (clusterselect_);
  vector <string> clusterreject = ResolveAliases (clusterreject_);

  if ((all || !jobs.empty() || !joblists.empty() || !status.empty()) &&
      clusters) {
    cerr << "ngstat: incompatible options" << endl;
    cerr << "Use \"ngstat -help\" for help" << endl;
    return 1;
  }

  int error = 0;
  ActivateGlobus();

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  // get user info

  CertInfo user;
  if (!user) {
    error = 1;
    goto errorexit;
  }

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft()) << endl;
  }

  // finding the information

  if (clusters) {

    vector <Giis> giislist;
    vector <Cluster> clusterlist;

    for (vector <string>::const_iterator vsi = clusterselect.begin();
	 vsi != clusterselect.end(); vsi++) {
      bool found = false;
      for (vector <Cluster>::iterator cli = clusterlist.begin();
	   !found && cli != clusterlist.end(); cli++)
	if (*vsi == cli->GetName()) found = true;
      if (!found) clusterlist.push_back (*vsi);
    }

    if (clusterlist.empty()) {
      if (GetGiises (giisurls, giislist)) {
	error = 1;
	goto errorexit;
      }

      clusterlist = FindClusters (giislist, user.GetSNx(), anonymous,
				  timeout, debug);
      if (clusterlist.empty()) {
	cerr << "ngstat: could not retrieve cluster list from giis" << endl;
	error = 1;
	goto errorexit;
      }
    }

    for (vector <string>::const_iterator vsi = clusterreject.begin();
	 vsi != clusterreject.end(); vsi++)
      for (vector <Cluster>::iterator cli = clusterlist.begin();
	   cli != clusterlist.end(); cli++)
	if (cli->GetName() == *vsi) {
	  if (debug) cout << "Rejecting cluster: " << *vsi << endl;
	  clusterlist.erase (cli);
	  break;
	}

    FindClusterInfo (clusterlist, Mds::ClusterInfo, user.GetSNx(), anonymous,
		     timeout, debug);

    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 cli != clusterlist.end(); cli++) {
      if (longlist)
	cli->Print();
      else
	cli->PrintShort();
      for (vector <Queue>::iterator qli = cli->queues.begin();
	   qli != cli->queues.end(); qli++) {
	if (longlist)
	  qli->Print();
	else
	  qli->PrintShort();
      }
      cout << endl;
    }
  }

  else {

    vector <string> jobids;
    vector <Cluster> clusterlist;

    jobids = GetJobIDs (jobs, joblists, clusterselect, clusterreject, status,
			all);

    if (all && jobids.empty()) {
      cout << "ngstat: No jobs" << endl;
      goto errorexit;
    }

    if (jobids.empty()) {
      cerr << "ngstat: no valid jobids given" << endl;
      cerr << "Use \"ngstat -help\" for help" << endl;
      error = 1;
      goto errorexit;
    }

    for (vector <string>::iterator vsi = jobids.begin();
	 vsi != jobids.end(); vsi++) {
      int pos0 = 0;
      if (vsi->substr (0, 9) == "gsiftp://") pos0 = 9;
      int pos = vsi->find_first_of (":/", pos0);
      if (pos == string::npos) {
	cerr << "ngstat: invalid jobid: " << *vsi << endl;
	error = 1;
	continue;
      }
      string clustername = vsi->substr (pos0, pos - pos0);
      bool found = false;
      for (vector <Cluster>::iterator cli = clusterlist.begin();
	   !found && cli != clusterlist.end(); cli++)
	if (clustername == cli->GetName()) found = true;
      if (!found) clusterlist.push_back (clustername);
    }

    if (clusterlist.empty()) {
      cerr << "ngstat: no valid jobids given" << endl;
      error = 1;
      goto errorexit;
    }

    FindClusterInfo (clusterlist, Mds::JobInfo, user.GetSNx(), anonymous,
		     timeout, debug);

    for (vector <string>::iterator vsi = jobids.begin();
	 vsi != jobids.end(); vsi++) {
      bool found = false;
      for (vector <Cluster>::iterator cli = clusterlist.begin();
	   !found && cli != clusterlist.end(); cli++)
	for (vector <Queue>::iterator qli = cli->queues.begin();
	     !found && qli != cli->queues.end(); qli++)
	  for (vector <Job>::iterator jli = qli->jobs.begin();
	       !found && jli != qli->jobs.end(); jli++)
	    if (*vsi == jli->GetId()) {
	      if (jli->IsStatus (status))
		if (longlist)
		  jli->Print();
		else
		  jli->PrintShort();
	      found = true;
	    }
      if (!found) {
	cerr << "ngstat: no job with jobid " << *vsi << " found" << endl;
	error = 1;
      }
    }
  }

 errorexit:
  DeactivateGlobus();
  return error;
}
